/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*  
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

/*
*
* ATmega128L Emulator
*
*/

/*
* Modifications:
*
* Copyright (c) 2004-2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*   * Redistributions of source code must retain the above copyright notice,
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its
*     contributors may be used to endorse or promote products derived from
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
*/


/*

  * Modification author: Hejun Wu (vmnet@cs.ust.hk)
  
********************************************************************************/

#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <libxml/parser.h>
#include <pthread.h>

#include "defs.h"
#include "atmega128.h"
#include "../VCPU/decode_tables.h"
#include "srec.h"
#include "iomanager.h"
#include "parse_config.h"
#include "table.h"
#include "elf_loader.h"

/* Cycle count to halt execution on */
uint64_t halttime = (uint64_t)(-1);

/* helper functions, globally needed */

// for logging in EM
///////////////////////////////////////////////////////////////////////////
unsigned short g_idChip=0;

unsigned short g_nChips =0;
////////////////////////////////////////////////////////////////////////////

int breakcomp(const void *a, const void *b){
    return (((break_t*)a)->addr - ((break_t*)b)->addr);
}

int watchcomp(const void *a, const void *b){
    watch_t *wa = (watch_t*)a, *wb = (watch_t *)b;
    /* return 0 if the ranges overlap, else return difference
	* between start addresses */
    if((wa->addr <= wb->addr) && (wa->addr + wa->length > wb->addr)){
        return 0;
    }
	
    if((wb->addr <= wa->addr) && (wb->addr + wb->length > wa->addr)){
        return 0;
    }
	
    return wa->addr - wb->addr;
	
}

void watchfree(void *a){
    watch_t *watch = (watch_t*)a;
	
    free(watch->value);
    free(watch);
}

/* note: this is only called by various "call" opcodes */
error_t atmega_push_help( atmega_chip *chip, uint32_t value )
{
    error_t result;
    uint16_t stack_index = PACK_INT_16(*chip->sph, *chip->spl);
    if(chip->pc_bits == 16){
        uint8_t vbb, vtb;
		
        vbb = LOW_BYTE_16( value );
        vtb = HIGH_BYTE_16( value );
        result = atmega_write_sram( chip, stack_index, vbb );
        stack_index--;
        if( result < 0 )
        {
            return result;
        }
        result = atmega_write_sram( chip, stack_index, vtb );
        stack_index--;
		
    }else{
        /* 22 bit PC */
        uint8_t vbb, vmb, vtb;
		
        vbb = LOW_BYTE_24( value );
        vmb = MID_BYTE_24( value );
        vtb = HIGH_BYTE_24( value );
        result = atmega_write_sram( chip, stack_index, vbb );
		
        stack_index--;
        if( result < 0 )
        {
            return result;
        }
        result = atmega_write_sram( chip, stack_index, vmb );
        stack_index--;
        if( result < 0 )
        {
            return result;
        }
        result = atmega_write_sram( chip, stack_index, vtb );
        stack_index--;
		
    }
	
    *chip->sph = HIGH_BYTE_16(stack_index);
    *chip->spl = LOW_BYTE_16(stack_index);
	
    chip->call_stack = stack_push( chip->call_stack, value );
    chip->stack_changed = 1;
	
    return result;
	
}

error_t atmega_pop_help( atmega_chip *chip, uint32_t *value )
{
    error_t result;
    uint16_t stack_index = PACK_INT_16(*chip->sph, *chip->spl);
    
    if(chip->pc_bits == 16){
        uint8_t vbb, vtb;
		
        stack_index++;
        result = atmega_read_sram( chip, stack_index, &vtb );
        if( result < 0 )
        {
            return result;
        }
		
        stack_index++;
        result = atmega_read_sram( chip, stack_index, &vbb );
		
        *value = PACK_INT_16( vtb, vbb );
    }else{
        /* 22 bit PC */
        uint8_t vbb, vmb, vtb;
		
        stack_index++;
        result = atmega_read_sram( chip, stack_index, &vtb );
        if( result < 0 )
        {
            return result;
        }
		
        stack_index++;
        result = atmega_read_sram( chip, stack_index, &vmb );
        if( result < 0 )
        {
            return result;
        }
		
        stack_index++;
        result = atmega_read_sram( chip, stack_index, &vbb );
		
        *value = PACK_INT_24( vtb, vmb, vbb );
    }
    
    *chip->sph = HIGH_BYTE_16(stack_index);
    *chip->spl = LOW_BYTE_16(stack_index);
	
    /* ret addr */
    chip->call_stack = stack_pop( chip->call_stack, NULL ); 
    chip->stack_changed = 1;
	
    return result;
}

void atmega_branch_skip_instruction( atmega_chip *chip )
{
    uint16_t instr;
    uint8_t t_byte, b_byte;
	
    instr=chip->flash[chip->pc+1];
    t_byte=TOP_BYTE(instr);
    b_byte=BOTTOM_BYTE(instr);
	
    if( (t_byte & 0xfe) == 0x90 && (b_byte & 0x0f) == 0x00){
        /* LDS */
        chip->pc += 3;
        chip->cycles += 3;
    }else if( (t_byte & 0xfe) == 0x92 && (b_byte & 0x0f) == 0x00){
        /* STS */
        chip->pc += 3;
        chip->cycles += 3;
    }else if( (t_byte & 0xfe) == 0x94 && (b_byte & 0x0e) == 0x0c){ 
        /* JMP */
        chip->pc += 3; 
        chip->cycles += 3;
    }else if( (t_byte & 0xfe) == 0x94 && (b_byte & 0x0e) == 0x0e){
        /* CALL */
        chip->pc += 3;
        chip->cycles += 3;
    }else{
        /* one-word instruction */
        chip->pc += 2;
        chip->cycles += 2;
    }
}

void atmega_alu_set_logic_flags( atmega_chip *chip, uint8_t R)
{
    CLR_FLAG( *chip->sreg, ATEMU_SREG_V );
    
    if ( R & 0x80 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_N );
        SET_FLAG( *chip->sreg, ATEMU_SREG_S );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_N );
        CLR_FLAG( *chip->sreg, ATEMU_SREG_S );
    }
    
    if ( R == 0x00 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
}

void atmega_alu_set_sub_flagsw( atmega_chip *chip, uint16_t R, uint16_t Rd)
{
    char n, v;
    uint16_t mask_15 = (uint16_t )1 << 15;
    
    /*Set V flag if two's complement overflow */
    if( (Rd & mask_15) & (~(R & mask_15)) )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_V);
        v = 0x1;
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_V );
        v = 0x0; 
    }
    
    /* set N if MSB of result is set */
    if ( R & mask_15 )
    {  
        SET_FLAG( *chip->sreg, ATEMU_SREG_N );
        n = 0x1;
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_N );
        n = 0x0;
    }
	
    /* set S for signedness */
    if ( v ^ n )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_S );
    }
    else 
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_S );
    }
	
    /*Set Z if the result is $0000, cleared otherwise*/
    if(R == 0x0000)
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    
    /*Set C if there was a carry from the MSB of the result*/
    if( (R & mask_15) & (~(Rd & mask_15)))
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_C );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_C );
    }    
}

void atmega_alu_set_add_flagsw( atmega_chip *chip, uint16_t R, uint16_t Rd)
{
    char n, v;
    uint16_t mask_15 = 1 << 15;
    
    /*Set V flag if two's complement overflow */
    if( (~(Rd & mask_15)) & (R & mask_15) )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_V);
        v = 0x1;
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_V );
        v = 0x0; 
    }
    
    /* set N if MSB of result is set */
    if ( R & mask_15 )
    {  
        SET_FLAG( *chip->sreg, ATEMU_SREG_N );
        n = 0x1;
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_N );
        n = 0x0;
    }
	
    /* set S for signedness */
    if ( v ^ n )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_S );
    }
    else 
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_S );
    }
	
    /*Set Z if the result is $0000, cleared otherwise*/
    if(R == 0x0000)
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    
    /*Set C if there was a carry from the MSB of the result*/
    if( (~(R & mask_15)) & (Rd & mask_15))
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_C );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_C );
    }    
}

void atmega_alu_set_add_flags( atmega_chip *chip, uint8_t R, uint8_t Rd, uint8_t Rr)
{
    char n, v;
    uint8_t mask_3 = 1 << 3;
    uint8_t mask_7 = 1 << 7;
	
    /* set H flag if borrow from bit 3 */
    if ( ( (Rd & mask_3) & (Rr & mask_3) ) ||
		( (Rr & mask_3) & (~(R & mask_3)) ) ||
		( (~(R & mask_3)) & (Rd & mask_3) ) )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_H );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_H );
    }
	
    /* set V if two's complement overflow from operation */
    if ( ( (Rd & mask_7) & (Rr & mask_7) & (~(R & mask_7)) ) ||
		( (~(Rd & mask_7)) & (~(Rr & mask_7)) & (R & mask_7) ) )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_V );
        v = 0x1;
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_V );
        v = 0x0;
    }
	
    /* set N if MSB of result is set */
    if ( R & mask_7 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_N );
        n = 0x1;
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_N );
        n = 0x0;
    }
	
    /* set S for signedness */
    if ( v ^ n )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_S );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_S );
    }
	
    /* set Z if result is 0 */
    if ( R == 0x00 )
		
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
	
    /* Set C if there was a carry from the MSB of the result */
    if ( ( (Rd & mask_7) & (Rr & mask_7) ) ||
		( (Rr & mask_7) & (~(R & mask_7)) ) ||
		( (~(R & mask_7)) & (Rd & mask_7) ) )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_C );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_C );
    }
}

void atmega_alu_set_sub_flags( atmega_chip *chip, uint8_t R, uint8_t Rd, uint8_t Rr)
{
    char n, v;
    uint8_t mask_3 = (uint8_t)1 << 3;
    uint8_t mask_7 = (uint8_t)1 << 7;
	
    /* set H flag if borrow from bit 3 */
    if ( ( (~(Rd & mask_3)) & (Rr & mask_3) ) ||
		( (Rr & mask_3) & (R & mask_3) ) ||
		( (R & mask_3) & (~(Rd & mask_3)) ) )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_H );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_H );
    }
	
    /* set V if two's complement overflow from operation */
    if ( ( (Rd & mask_7) & (~(Rr & mask_7)) & (~(R & mask_7)) ) ||
		( (~(Rd & mask_7)) & (Rr & mask_7) & (R & mask_7) ) )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_V );
        v = 0x1;
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_V );
        v = 0x0;
    }
	
    /* set N if MSB of result is set */
    if ( R & mask_7 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_N );
        n = 0x1;
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_N );
        n = 0x0;
    }
	
    /* set S for signedness */
    if ( v ^ n )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_S );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_S );
    }
	
    /* previous value remains unchanged if 0 */
    if ( R == 0x00 )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_Z );
    }
	
    /* Set C if absolute R of K is larger than absolute R of Rd */
    if ( ( (~(Rd & mask_7)) & (Rr & mask_7) ) ||
		( (Rr & mask_7) & (R & mask_7) ) ||
		( (R & mask_7) & (~(Rd & mask_7)) ) )
    {
        SET_FLAG( *chip->sreg, ATEMU_SREG_C );
    }
    else
    {
        CLR_FLAG( *chip->sreg, ATEMU_SREG_C );
    }
}
/*************************************************/
/*************************************************/

/**
* Static Functions 
*/

error_t atmega_initialize_chip( atmega_chip *chip )
{
    if ( chip == NULL )
		
    {
        return -EFAULT;
    }
	
    chip->sram = (uint8_t *) malloc( sizeof(uint8_t) * chip->sram_bytes );
	
    if ( chip->sram == NULL )
    {
        return -EFAULT;
    }
	
    memset( chip->sram, 0x00, chip->sram_bytes );
	
    /**
	* We have to load an image after this.  Set 'flash' to NULL now to 
	* avoid executing anything if we forget to load our program.
	*/
    chip->flash = NULL;
    chip->pc = 0;
	
    chip->sreg = chip->sram + chip->sreg_io;
	
    chip->eind = chip->sram + chip->eind_io;
	
    chip->register_file = chip->sram;
    
    chip->spl = chip->sram + chip->spl_io;
    chip->sph = chip->sram + chip->sph_io;
	
    chip->rampz = chip->sram + chip->rampz_io;
	
    chip->cycles = 0;
    chip->sh = 0;
	
    chip->mode=ATEMU_MODE_EXECUTE;
	
    chip->xmlid = NULL;
	
    return 0;
}

uint16_t atmega_fetch_op_addr( atmega_chip *chip )
{
    return CONVERT_ADDR(chip->flash[chip->pc + 1]);
}

error_t atmega_read_sram( atmega_chip *chip, uint16_t addr,
						 uint8_t *value )
{
    if(addr == chip->sreg_io ||
		addr == chip->spl_io ||
		addr == chip->sph_io ||
		addr == chip->rampz_io){
        /* special cases, in IO space, not IO registers */
        *value = chip->sram[addr];
    } else if(addr >= chip->io_offset && addr <= chip->io_last){
        atmega_ioreg_read(chip, addr, value);
    }else if(addr < chip->sram_bytes){
        *value = chip->sram[addr];
    }else{
        log_error(chip, "Access outside of SRAM.\n");
        return -1;
    }
    return 0;
}
unsigned char g_tempMemDbg=0;
error_t atmega_write_sram( atmega_chip *chip, uint16_t addr, uint8_t value )
{
    chip->sram[addr] = value;
    if(addr >= chip->io_offset && addr <= chip->io_last &&
        /* special cases, in IO space, not IO registers */
        !( addr == chip->sreg_io ||
        addr == chip->spl_io ||
        addr == chip->sph_io ||
        addr == chip->rampz_io))
    {
        atmega_ioreg_write(chip, addr, value);
		
    }else if(addr < chip->sram_bytes){
        watch_t *wpoint, query;
		
        query.addr = addr;
        query.length = 1;
        wpoint=table_find(chip->watchtable, &query);
        if(wpoint != NULL && wpoint->value[addr - wpoint->addr] != value){
            log_debug(chip, V_QUIET, "watchpoint 0x%x hit.\n", wpoint->addr);
            chip->watchhit=1;
            wpoint->value[addr - wpoint->addr] = value;
        }
		
    }else{
        log_error(chip, "Access outside of SRAM.\n");
        return -1;
    }
	
    return 0;
}

void atmega_decode( atmega_chip *chip )
{

    uint8_t enc;
	
    chip->t_byte = TOP_BYTE(chip->instr);
    chip->b_byte = BOTTOM_BYTE(chip->instr);
	
    chip->th_nibble = HIGH_NIBBLE(chip->t_byte);
    chip->tl_nibble = LOW_NIBBLE(chip->t_byte);
    chip->bh_nibble = HIGH_NIBBLE(chip->b_byte);
    chip->bl_nibble = LOW_NIBBLE(chip->b_byte);
	
    enc = encode_table[chip->t_byte];
	//VMNet power logging in EM+++++++++++++++++++++++++++

#ifdef POWER_CONSUMPTION_MEASURE
	{
		uint8_t isActive=1;
		if(chip->t_byte==0x95) //sleep
		{
			//if((chip->bh_nibble==0x8)&&(chip->tl_nibble!=0x4))
			if(chip->b_byte==0x88)
				isActive=0;
		}
		
		if(isActive)
			LOG_CURRENT_OPERATION(CPU_ACTIVE,CPU_MASK);
	}
#endif   
  
                                     
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++
    switch( enc )
    {
	case 0x0:
		break;
		
	case 0x1:
		chip->reg_index_a = ( chip->bh_nibble & 0x7 ) + 16;
		chip->reg_index_b = ( chip->bl_nibble & 0x7 ) + 16;
		
		chip->a = chip->register_file[chip->reg_index_a];
		chip->b = chip->register_file[chip->reg_index_b];
		break;
		
	case 0x2:
		chip->reg_index_a = chip->bh_nibble + 16;
		chip->reg_index_b = chip->bl_nibble + 16;
		
		chip->a = chip->register_file[chip->reg_index_a];
		chip->b = chip->register_file[chip->reg_index_b];
		break;
		
	case 0x3:
		chip->reg_index_a = chip->bh_nibble | (( chip->tl_nibble & 1 ) << 4);
		chip->reg_index_b = chip->bl_nibble | (( chip->tl_nibble & 2 ) << 3);
		
		chip->a = chip->register_file[chip->reg_index_a];
		chip->b = chip->register_file[chip->reg_index_b];
		break;
		
	case 0x4:
		chip->reg_index_a = chip->bh_nibble + 16;
		
		chip->a = chip->register_file[chip->reg_index_a];
		chip->b = (chip->tl_nibble << 4) | (chip->bl_nibble);
		break;
		
	case 0x5:
		chip->reg_index_a = chip->bh_nibble | (( chip->tl_nibble & 1 ) << 4);
		
		chip->a = chip->register_file[chip->reg_index_a];
		break;
		
	case 0x6:
		chip->reg_index_a = chip->bh_nibble | (( chip->tl_nibble & 1 ) << 4);
		
		chip->a = chip->register_file[chip->reg_index_a];
		chip->b = ( chip->bl_nibble & 0x7 ) | 
			( ( chip->tl_nibble << 1 ) & 0x18 ) | 
			( ( chip->th_nibble << 4 ) & 0x20 );
		break;
		
	case 0x7:
		chip->reg_index_a = chip->bh_nibble | (( chip->tl_nibble & 1 ) << 4);
		
		chip->a = chip->register_file[chip->reg_index_a];
		
		if ( chip->bl_nibble == 0x0 )
		{
			chip->addr = atmega_fetch_op_addr( chip );
		}
		
		break;
		
	case 0x8:
		if( chip->bl_nibble == 0x8 )
		{
			chip->a = chip->bh_nibble & 0x7;
		}
		else if ( (chip->bl_nibble & 0xc ) == 0xc )
		{
			chip->addr = atmega_fetch_op_addr( chip );
			
			/* For the extra address bits of JMP and CALL.  These bits 
			allow the JMP and CALL to go beyond the 16-bit range... we 
			will never need this... */
			
			chip->addr |= ((uint32_t) ( chip->bl_nibble & 0x1 )) << 16;
			chip->addr |= ((uint32_t) chip->bh_nibble) << 17;
			chip->addr |= ((uint32_t) ( chip->tl_nibble & 0x1 )) << 21;
		}
		else
		{
			chip->reg_index_a = chip->bh_nibble | (( chip->tl_nibble & 1 ) << 4);
			
			chip->a = chip->register_file[chip->reg_index_a];
		}
		
		break;
		
	case 0x9:
		chip->reg_index_a = chip->bh_nibble << 1;
		chip->reg_index_b = chip->bl_nibble << 1;
		
		chip->a = chip->register_file[chip->reg_index_a];
		chip->b = chip->register_file[chip->reg_index_b];
		break;
		
	case 0xa:
		chip->reg_index_a = ( (chip->bh_nibble & 0x3) << 1 ) + 24;
		
		chip->a = chip->register_file[chip->reg_index_a];
		chip->b = chip->bl_nibble | ( ( chip->bh_nibble  & 0xc ) << 2);
		break;
		
	case 0xb:
		chip->a = (chip->bh_nibble << 1) | ((chip->bl_nibble & 0x8) >> 3);
		chip->b = chip->bl_nibble & 0x7;
		break;
		
	case 0xc:
		chip->reg_index_a = chip->bh_nibble | (( chip->tl_nibble & 1 ) << 4);
		
		chip->a = chip->register_file[chip->reg_index_a];
		chip->b = chip->bl_nibble | (( chip->tl_nibble & 6 ) << 3);
		break;
		
	case 0xd:
		chip->addr = CONVERT_ADDR(chip->instr) & 0x0fff;
		break;
		
	case 0xe:
		chip->a = (chip->bl_nibble >> 3) & 0x1;
		chip->a |= chip->bh_nibble << 1;
		chip->a |= (chip->tl_nibble & 0x3) << 5;
		chip->b = chip->bl_nibble & 0x7;
		break;
		
	case 0xf:
		chip->reg_index_a = chip->bh_nibble | 
			(( chip->tl_nibble & 0x1) << 4 );
		
		chip->a = chip->register_file[chip->reg_index_a];
		chip->b = chip->bl_nibble & 0x7;
		break;
		
	default:
		log_error(chip,  "Unknown encoding." );
    }

}

/*************************************************/
/*************************************************/

static inline int iself(char *buffer, size_t bytes) {
    return (memcmp(buffer, "\177ELF", (bytes > 4) ? 4 : bytes) == 0);
}

static inline int issrec(char *buffer, size_t bytes) {
    return (memcmp(buffer, "S0", (bytes > 2) ? 2 : bytes) == 0);
}

static int isoldrunfile(FILE *file) {
    char buffer[256];
    int  foundline = 0;
	
    /* The SREC/ELF search had caused a seek from the beginning */
    fseek(file, 0, SEEK_SET);
	
    /* Look for a line not blank or beginning with # */
    while(fgets(buffer, sizeof(buffer), file)) {
        if(!isspace(buffer[0]) && (buffer[0] != '#') && (buffer[0] != ';')) {
            foundline = 1;
            break;
        }
    }
	
    /* If a line is not found, or this line does not begin with NCHIPS,
	* this is not a run file */
    if(!foundline || (strncmp(buffer, "NCHIPS", 6) != 0)) {
        return 0;
    } else {
        return 1;
    }
}

static int isnewrunfile(FILE *file) {
    char buffer[4];
    size_t bytes;
	
    /* The SREC/ELF search had caused a seek from the beginning */
    fseek(file, 0, SEEK_SET);
	
    /* If the first four bytes contain "<?xm" in some encoding, we have
	* an XML file (close enough for our purposes) */
    bytes = fread(buffer, 1, 4, file);
    if(bytes == 4) {
        if(memcmp(buffer, "<?xm", 1) == 0) {
            /* UTF-8, ASCII, ISO-8859-1, etc. */
            return 1;
        } else if(memcmp(buffer, "<\000?\000", 2) == 0) {
            /* UTF-16 LE */                  
            return 1;
        } else if(memcmp(buffer, "\000<\000?", 2) == 0) {
            /* UTF-16 BE */
            return 1;
        } else if(memcmp(buffer, "<\000\000\000", 4) == 0) {
            /* UCS4 LE */
            return 1;
        } else if(memcmp(buffer, "\000\000\000<", 4) == 0) {
            /* UCS4 BE */
            return 1;
        } else if(memcmp(buffer, "\357\273\277<", 4) == 0) {
            /* UTF-8 with Byte Order Mark */
            return 1;
        } else if(memcmp(buffer, "\377\376<\000", 4) == 0) {
            /* UTF-16 LE with Byte Order Mark */
            return 1;
        } else if(memcmp(buffer, "\376\377\000<", 4) == 0) {
            /* UTF-16 BE with Byte Order Mark */
            return 1;
        } else if(memcmp(buffer, "\376\377\000\000", 4) == 0) {
            /* UCS4 LE with Byte Order Mark */
            return 1;
        } else if(memcmp(buffer, "\000\000\377\376", 4) == 0) {
            /* UCS4 BE with Byte Order Mark */
            return 1;
        }
    }
	
    return 0;
}


/**
* Public Functions
*/

int atemu_load_file( atmega_chip ***chips, char  *fname, char *config )
{
    int nchips = 0;
    int nret;
    /* Variables for signature testing */
    FILE *file;
    char input[4];
    size_t bytes;
#if defined(MAXPATHLEN)
    char filename[MAXPATHLEN];
#else
    char filename[256];
#endif
	
    /* Get the first couple bytes of the input file */
    file = atmega_findfile(fname, filename, sizeof(filename), TYPE_RUNFILE);
    if(file == NULL) {
        *chips = NULL;
        return 0;
    }
	
    /* Check if it is an ELF or an SREC */
    bytes = fread(input, 1, 4, file);
    if((bytes > 0) && (iself(input, bytes) || issrec(input, bytes))) {
        fclose(file);
		
        /* Display some messages */
        printf("Number of sensor motes: 1\n");
        printf("Loading ELF \"Chip 0\" (1 of 1)\n");
		
        /* Attempt to load the ELF/SREC */
        *chips = malloc( sizeof(atmega_chip *) );
		
        if((*chips, 0, config, filename) < 0) {
            free(*chips);
            *chips = NULL;
            return 0;
        }
		
        fflush(stdout);
        return 1;
    }
	
    /* Check if it is a runfile */
    if(isoldrunfile(file)) {
        fclose(file);
        *chips = NULL;
        fprintf(stderr, "Error: %s is in the old  run file format.  Please"
			" convert it.\n", fname);
        return 0;
    }
	
    /* Check if a new runfile (XML) */
    if(isnewrunfile(file)) {
        fclose(file);
        nret=  atmega_parse_runfile(chips, &nchips, filename);
		// VMNet for testing no use...++++++++++++++++++++++++++++++++
        if(nret < 0) {
            *chips = NULL;
			if(nret==-2)
				nchips=-2;
			else
				nchips = 0;
        }                                               
		//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        return nchips;
    }
	
	
    /* Unknown file type */
    fclose(file);
    fprintf(stderr, "Error: %s is in an unknown file format.\n", fname);
    return 0;
}

error_t atmega_load_chip(atmega_chip **chips, int i, char *config, char *srec) {
    error_t result;
    device_obj *devhead;
#ifdef MAXPATHLEN
    char model[MAXPATHLEN];
    char flash[MAXPATHLEN];
#else
    char model[256]={0};
    char flash[256]={0};
#endif
    FILE *file;
    int idMote;
	
    /* Locate configuration file */
    file = atmega_findfile(config, model, sizeof(model), TYPE_CONFFILE);
    if(file == NULL) {
        return -ENOENT;
    }
    fclose(file);
	
    /* Locate flash */
    if(srec[0] != '/') {
        file = atmega_findfile(srec, flash, sizeof(flash), TYPE_FLASH);
        if(file == NULL) {
            return -ENOENT;
        }
        fclose(file);
    } else {
        strncpy(flash, srec, sizeof(flash));
        flash[sizeof(flash) - 1] = 0;
    }
	
	// printf("\tModel: %s\n\tFlash: %s\n", model, flash);
    idMote=i;
    /* Create the Chip */
    g_idChip=i;

    result = atmega_create_chip( &(chips[i]), model );
    if(result < 0) {
        return result;
    }
	
    chips[i]->id = idMote;

    result = atmega_load_flash(chips[i], flash);
	
    if(result < 0) {
        fprintf(stderr, "\tError: Failed to load flash.\n");
        atmega_destroy_chip( &(chips[i]) );
        return result;
    }
	
    /* Notify the devices loading has completed */
    devhead = chips[i]->device_head;
    while(devhead) {
        if(devhead->wake) {
            devhead->wake(devhead);
        }
		
        devhead = devhead->next;
    }
	          
    return 0;
}

/**
* atmega_create_chip
*
* Creates a structure to store the ATmega128L processor
* state. 
*
*/

error_t atmega_create_chip( atmega_chip **ret_chip, char *conf )
{
    error_t result;
    atmega_chip *chip = NULL;
	
    if(ret_chip == NULL) {
        return -EFAULT;
    }
	
    chip = (atmega_chip *) calloc( sizeof(atmega_chip), 1 );
	
    /* Create the chip itself */
	
    chip = (atmega_chip*)calloc(sizeof(atmega_chip), 1);
	
    if(chip == NULL) {
      exit(1);
        return -ENOMEM;
    }
	
    /* Parse the chip configuration */
    result = atmega_parse_config(chip, conf);
	
    if(result < 0) {
        *ret_chip = NULL;
        atmega_destroy_chip(&chip);
        exit(1);
        return result;
    }
	
    /* Initialize Processor State */
    result = atmega_initialize_chip(chip);
	
    if(result < 0) {
        *ret_chip = NULL;
        atmega_destroy_chip(&chip);
                exit(1);
        return result;
    }
	
    /* Initialize the breakbpoint table */
    chip->breaktable=table_create(breakcomp);
    if(chip->breaktable == NULL){
        fprintf(stderr, "\tError: Failed to allocate breakpoint table\n");
        return -1;
    }
	
    /* Initialize the watchbpoint table */
    chip->watchtable=table_create(watchcomp);
    if(chip->watchtable == NULL){
        fprintf(stderr, "\tError: Failed to allocate watchpoint table\n");
        return -1;
    }
	
    /* Initialize the notifypoint table */
    chip->notifytable=table_create(watchcomp);
    if(chip->notifytable == NULL){
        fprintf(stderr, "\tError: Failed to allocate notifypoint table\n");
        return -1;
    }
	
    chip->call_stack = NULL;
	
    /* Return created chip */
    *ret_chip = chip;
	
    return 0;
}

/**
* atmega_destroy_chip
*
* Frees any memory used by the ATmega128L state structure.
*
*/
error_t atmega_destroy_chip( atmega_chip **chip )
{
    if ( ( chip == NULL ) || ( *chip == NULL ) )
    {
        return -EFAULT;
    }
	
    if((*chip)->sram != NULL){
        free((*chip)->sram);
    }
	
    if((*chip)->flash != NULL){
        free((*chip)->flash);
    }
	
    atmega_ioman_cleanup(*chip);
	
    if((*chip)->breaktable != NULL){
        table_destroy((*chip)->breaktable, free);
    }
    
    if((*chip)->watchtable != NULL){
        table_destroy((*chip)->watchtable, watchfree);
    }
	
    if((*chip)->notifytable != NULL){
        table_destroy((*chip)->notifytable, free);
    }
	
    if((*chip)->xmlid != NULL) {
        xmlFree((xmlChar*)((*chip)->xmlid));
    }
	
    free( *chip );
    *chip = NULL;
	
    return 0;
} 
////////////////////////////////////////////////////////////////////////////////
// VMNet: find a pattern for post parse flash to add attributes
////////////////////////////////////////////////////////////////////////////////

int pattern_matching_flash(uint8_t* buffer, uint8_t *pattern, int patternlen, int flashlen)
{
	int i=0,j=0;
	while(i<=flashlen)
	{
		if(buffer[i]==pattern[j])
		{
			while(i<=flashlen)
			{
				if(buffer[i++]!=pattern[j++])
				{
					j=0;
					break;
				}
				else
				{
					if(j==patternlen)
						return i;
				}
			}
		}
		i++;
	}
	return -1;
}
/////////////////////////////////////////////////////////////////////////////////
//VMNet post parse the flash. to set the position of virtual motes
//enable query position of motes
/////////////////////////////////////////////////////////////////////////////////
void post_set_position(atmega_chip *chip, uint16_t x, uint16_t y,uint16_t z)
{
  int i=0;
  uint8_t * readBuffer=(uint8_t *)chip->flash;
  
  int flashlen =sizeof(uint16_t) * chip->flash_words;
  i=pattern_matching_flash(readBuffer, "whj_position_x", 14,flashlen);
  
  if(i>=0)
  {

    readBuffer[i]=(unsigned char)(x&0xff);
		readBuffer[i+1]=(unsigned char)((x&0xff00)>>8);
  }
  else return;
  i=pattern_matching_flash(readBuffer, "whj_position_y", 14,flashlen);
  if(i>=0)
  {

    readBuffer[i]=(unsigned char)(y&0xff);
 		readBuffer[i+1]=(unsigned char)((y&0xff00)>>8);
  }
  else
  return;
}       
////////////////////////////////////////////////////////////////////////////////
// VMNet: automatically changes the local address in motes
// Parse the executable binary code to add ID of mote.
// post_parse_flash
////////////////////////////////////////////////////////////////////////////////

void post_parse_flash(atmega_chip *chip)
{
	char foundflag=0;
	int i=0;
	uint8_t * readBuffer=(uint8_t *)chip->flash;
	int flashlen =sizeof(uint16_t) * chip->flash_words;
	{	
	i=pattern_matching_flash(readBuffer, "nodeid", 6,flashlen);
	i-=10;
	if(i>0)
	{
		readBuffer[i]=(unsigned char)(g_idChip&0xff);
		readBuffer[i+1]=(unsigned char)((g_idChip&0xff00)>>8);
	}
	else
	{
		printf("Fatal error: Not found nodeid!!\n");
		exit(1);
	}
	}  
	#ifdef NEW_VERSION_NODEID
	{
		i=pattern_matching_flash(readBuffer, "whj_nodeid", 10,flashlen);
		if(i>=0)
		{

			readBuffer[i]=(unsigned char)(g_idChip&0xff);
			readBuffer[i+1]=(unsigned char)((g_idChip&0xff00)>>8);
		}
		else
		{
			printf("Fatal error: Not found new version nodeid!!\n");
			exit(1);
		}
	}
	#endif
	
	

}  
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//---------------------------------------------------------------------
//VMNet modify this
/**
* atmega_load_flash
*
* Loads the chips flash memory with the specified image.
* 
*/
//////////////////////////////////////////////////////////////////////////////////////////////////////
error_t atmega_load_flash( atmega_chip *chip, char *filename )
{
    error_t result;
    chip->flash = (uint16_t *) malloc( sizeof(uint16_t) * chip->flash_words );
	
    if ( chip->flash == NULL )
    {
        return -ENOMEM;
    }

    memset(chip->flash, 0, sizeof(uint16_t) * chip->flash_words );
    result = elf_load( (uint8_t *) chip->flash, 
        sizeof(uint16_t) * chip->flash_words, &chip->pc,
        filename );
	
    if ( result < 0 )
    {
        result = srec_read( (uint8_t *) chip->flash,
            sizeof(uint16_t) * chip->flash_words, &chip->pc,
            filename );
		
        if ( result < 0 )
        {
            printf("loading error in file format\n");
            free( chip->flash );
            chip->flash = NULL;
            return result;
        }
    }     
	//VMNet: add the local address here. i.e. the ID of mote
    post_parse_flash(chip);      
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++
    return 0;
}

/**
* atmega_step
*
* Completes one instruction.
*
*/
#if 1
#define MAX_RANDOM_WAIT 17777777
#define INITAILIZATION_SECONDS (1)
#define PATCH_CYCLES (10000)
#define TIME_UNIT_CYCLES (737280)
static char flagLog=0;
static char flagQuery=0;
static char flagES=0;
char g_isQueryStarted = 0;

#endif 
void start_query()
{

	system(g_query_str);      
	
}




error_t atmega_step( atmega_chip *chip)
{
    int interrupt, reg;
    error_t result = 0;
    // for loging in VMNet+++++++++++++++++++++++
   
    /* if decoding (to screen) skip all the rest */
//VMNet random starts the VMs. to enable them not starts simultaneously
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   g_idChip=chip->id;
   
   
#ifdef EMULATION_SPEED_MEASURE
	
   	if(!flagES)
	{
		if(!g_startCycles)
	 	g_startCycles=g_chips[0]->cycles;
		else
		if((g_chips[0]->cycles/7372800-g_startCycles/7372800)>=(uint32_t)EMULATION_SPEEDTEST_START_SECONDS)
		{
			flagES =1;
			//RECORD_START_TIME();
			START_POWER_LOGGING();
		}
	}
	else
	if((g_chips[0]->cycles/7372800-g_startCycles/7372800)>=(uint32_t)EMULATION_SPEEDTEST_STOP_SECONDS)
	{
		fprintf(stderr,"finished logging, auto exiting... Remember to copy the logs!!\n");
		dump_power_data_log();
	}
#else 
#ifdef EPOCH_POWER_MEASURE
   	if(!flagES)
	{
		if(!g_startCycles)
		{
	 	g_startCycles=g_chips[0]->cycles;
		}
		else
		if((g_chips[0]->cycles/7372800-g_startCycles/7372800)>=(uint32_t)LOG_CONTROL_START)
		{
			flagES =1;
			//RECORD_START_TIME();
			if(g_query_id)
				g_start_logs = 1;
			else
				START_POWER_LOGGING();
		}
	}
	else
	if((g_chips[0]->cycles/7372800-g_startCycles/7372800)>=(uint32_t)LOG_CONTROL_END)
	{
		fprintf(stderr,"finished logging, auto exiting... Remember to copy the logs!!\n");
		dump_power_data_log();
		
	}

#endif

#endif
 #ifdef BISQUE_TEST
	if((!flagQuery)&&(chip->cycles/TIME_UNIT_CYCLES>0))   
	{
	   pthread_t start_query_thread;
	   pthread_create(&start_query_thread, NULL, start_query, NULL);
	  // benchmark_send_query(TEST_QUERY_ID);
		flagQuery=1;		
		
	}
	#endif
   #if 0
  // whj test... if(!flagLog)
   {
   	
   	if((!flagQuery)&&(chip->cycles/TIME_UNIT_CYCLES>0))   
	{
	   pthread_t start_query_thread;
	   pthread_create(&start_query_thread, NULL, start_query, NULL);
	   benchmark_send_query(TEST_QUERY_ID);
		flagQuery=1;		
		
	}
	
	#if 0
	else if((flagQuery)&&(chip->cycles/TIME_UNIT_CYCLES>(INITAILIZATION_CYCLES+PATCH_CYCLES)))
	{
		flagLog=1;
		g_log_startreg_flag = 1;
		g_startCycles=g_chips[0]->cycles;
		g_isQueryStarted = 1;
		printf("send query...\n");
		benchmark_send_query(g_query_id);
		
	}
	#endif
	
   }
   
  
#endif
    //VMNet
    // random start virtual motes++++++++++++++++
    #if 0
    if(g_number_start<g_nChips)
    {
      if(g_idChip){
      if(!g_current_waited_cycles[g_idChip])
      {
          
          g_current_waited_cycles[g_idChip]=random() % MAX_RANDOM_WAIT;
          printf("%d,current_waited_cycles =%d\n",g_idChip, g_current_waited_cycles[g_idChip]);
        chip->cycles=g_chips[0]->cycles;           
          return 0;
      }
      else
      {
        chip->cycles=g_chips[0]->cycles;
        g_current_waited_cycles[g_idChip]--;
        if(!g_current_waited_cycles[g_idChip])
         {
           g_current_waited_cycles[g_idChip]--;
           g_number_start++;
           if(g_number_start==(g_nChips-1))
                printf("Initialization Done! Instructions running!\n");

         }
        else
          return 0;
      }
      }
    }
    #endif
    //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#if 0
    if(chip->mode == ATEMU_MODE_DECODE){
        chip->instr = chip->flash[chip->pc];
        atmega_decode( chip );
        return (*(instruction_table[chip->t_byte]))( chip );
    }
#endif
	
    if(!chip->am_sleeping) {
        chip->oldpc=chip->pc;
        chip->oldcycles=chip->cycles;
		
        /* Fetch */
        chip->instr = chip->flash[chip->pc];
		
        /* Decode/Register Fetch */
        atmega_decode( chip );
		
        /* Execute */
        /* Memory Access */
        /* Write Back */
        result=(*(instruction_table[chip->t_byte]))( chip );
        /* help reduce errant array access */
        if(chip->pc > chip->flash_words){
            chip->pc %= chip->flash_words;
            log_error(chip, "Warning: PC pointed to outside the program"
				" space\n");
        }
    } else {
        /* Keep the emulator happily ticking away */
        chip->cycles++;
    }
	
    /* Poll for interrupts */
    if((chip->interrupts != 0) && GET_FLAG(*chip->sreg, ATEMU_SREG_I)) {
        if(chip->am_sleeping) {
            /* We need total of four cycles to wake up */
            chip->cycles += 3;
            chip->am_sleeping = 0;
            atmega_ioman_leave_sleep(chip);
        } else {
            interrupt = atmega_ioman_poll_interrupts(chip);
            if(interrupt != -1) {
                /* Interrupt occured: disable interrupts, call handler */
		#if 0
                if (logfile && logintr) {
                    fprintf(logfile, "%03d INTERRUPTION %15ld\t%x\t%d\n", chip->id, (long)chip->cycles,
						(int)chip->pc, interrupt);
                }
		#endif
				//  printf("%03d INTERRUPTION %15ld\t%x\t%d\n", chip->id, (long)chip->cycles,
                //            (int)chip->pc, interrupt);
				
                CLR_FLAG(*chip->sreg, ATEMU_SREG_I);
                atmega_push_help(chip, chip->pc);
                chip->pc = ((uint32_t)interrupt) * 2 - 2;
            }else{
                log_error(chip, "chip->interrupts != 0, but no device"
					" interrupted us.\n");
            }
        }
    }
	//VMNet remove. VMNet has log+++++++++++++++++++++++++++++++++++++++++++++++
#if 0
    if (logfile && logstack) {
		// check for and log stack changes /
		if(chip->sh != PACK_INT_16(*chip->sph, *chip->spl)) {
			chip->sh = PACK_INT_16(*chip->sph, *chip->spl);
			fprintf(logfile, "%03d STAK %15ld\t%x\t%x\t%d\n", chip->id, (long)chip->cycles, 
				(int)chip->pc, (int)chip->sh, 0x10ff - (int)chip->sh);
		}
    }
    if (logfile && logmachine && !chip->am_sleeping) {
		//  randomly dump machine state /
		if (lrand48() % 1000 == 0) {
			fprintf(logfile, "%03d DUMP %15ld\t%x\t%x\t%x", chip->id, (long)chip->cycles, 
				(int)chip->pc, PACK_INT_16(*chip->sph, *chip->spl), 
				(int)*chip->sreg);		  
			for(reg = 0; reg < 32; reg++ ) {
				fprintf(logfile, "\t%x", chip->register_file[reg] );
			}
			fprintf(logfile, "\n");
		}
    } ///
#endif
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    return result;
}

