/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
* 
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

/*
* Modifications:
*
* Copyright (c) 2004-2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*   * Redistributions of source code must retain the above copyright notice,
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its
*     contributors may be used to endorse or promote products derived from
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
*
*/


/*

  * Modification author: Hejun Wu (vmnet@cs.ust.hk)
  
********************************************************************************/


#include <assert.h>
#include <string.h>
#include <dlfcn.h>
#include <search.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <libxml/parser.h>
#include <libxml/xmlerror.h>

#include "iomanager.h"
#include "debug.h"
#include "table.h"
#include "parse_config.h"
#include "atmega128.h"
#define ADD_LIST(x,h) x->next=h; h=x;
#define RUNDIR "run"
typedef struct{
    void *dlhandle;
    ioman_create_func *create_device;
    ioman_register_func *add_register;
    ioman_interrupt_func *add_interrupt;
    ioman_clock_func *add_clock;
    ioman_param_func *set_param;
    ioman_uses_func *set_device;
}device_funcs;

static const char *so_paths[] = {
    "./devices/%s."SOEXT,   /* look relative to self (build directory) */
		DEVICEDIR"/%s."SOEXT,   /* or here (installed directory) */
		"%s."SOEXT,             /* as a last stab, library search path */
		NULL
};

static const char *conf_paths[] = {
    /* Default configuration paths */
    "%s",
		"config/%s",
		CONFDIR"/%s",
		/* Default configuration paths (user didn't append extension) */
		"%s.xconf",
		"config/%s.xconf",
		CONFDIR"/%s.xconf",
		NULL
};

static const char *run_paths[] = {
    /* Normal run files */
    "%s",
		"run/%s",
		RUNDIR"/%s",
		/* Normal run files (user didn't provide extension) */
		"%s.xrun",
		"run/%s.xrun",
		RUNDIR"/%s.xrun",
		/* A flash dump (ELF exe or SREC) */
		"%s.exe",
		"%s.srec",
		NULL
};

static const char *flash_paths[] = {
    "%s",
		"%s.exe",
		"%s.srec",
		NULL
};

/* Namespace for configuration files */
static const char namespace[] = "http://www.cshcn.umd.edu/research/atemu/";

/* I/O Register Table (might be temporary) */
#define PARSER_MAX_REGS 256

typedef struct {
    uint8_t reg;
    char *name;
} io_reg;
static table_t *io_reg_table = NULL;

/* If a XML node is an element and of the desired namespace and name,
* returns true.  Returns false otherwards. */
int IsElement(xmlNodePtr node, const char* namespace, const char* nodename) {
    assert(node != NULL);
	
    /* Make sure the node is an element */
    if(node->type != XML_ELEMENT_NODE) {
        return 0;
    }
	
    /* Make sure of the correct namespace */
    if(namespace) {
		
        if(node->ns && (strcmp((char*)node->ns->href, namespace) != 0)) {
            return 0;
        }
    }
	
    /* Make sure the correct element */
    if(nodename) {
        if(strcmp((char*)node->name, nodename) != 0) {
            return 0;
        }
    }
	
    /* Everything's good */
    return 1;
}

/*
* Locate various program datafiles by iterating through the available
* search paths.  Will return a file handle (read-only from fopen) if

  * can be found and will copy the filename into fileout.  If cannot be
  * found, sets fileout to an empty string and returns NULL.
  *
  * type effects what search path to use:
  * TYPE_RUNFILE  - Searchs for a run file (mote list)
  * TYPE_CONFFILE - Searchs for a configuration file
  * TYPE_FLASH    - Searchs for a flash image (elf or srec)
*/
FILE* atmega_findfile(const char* filename, char* fileout, size_t filemax,
					  int type) {
    const char **pathlist;
    char *eol;
    FILE *file;
	
    /* Check parameters */
    assert(filename != NULL);
    assert(fileout != NULL);
    assert(filemax > 0);
	
    /* Select which set of paths to search */
    switch(type) {
    case TYPE_RUNFILE:
        /* Looking for a run file */
        pathlist = run_paths;
        break;
    case TYPE_CONFFILE:
        /* Looking for a configuration file */
        pathlist = conf_paths;
        break;
    case TYPE_FLASH:
        /* Looking for a flash (elf or srec) */
        pathlist = flash_paths;
        break;
    default:
        /* ?! */
        assert(0);
        return NULL;
    }
	
    /* Don't search if already absolute */
    if((filename[0] == '/')
#ifdef WIN32
		|| (filename[0] == '\\')
		|| (isalpha(filename[0]) && (filename[1] == ':'))
#endif
		) {
        /* Copy the filename and attempt to open it */
        strncpy(fileout, filename, filemax);
        fileout[filemax - 1] = 0;
        file = fopen(fileout, "r");
        if(file != NULL) {
            return file;
        }
        /* Display the appropriate filename */
        if(errno == ENOENT) {
            fprintf(stderr, "Error: Could not find %s.\n", filename);
        } else {
            fprintf(stderr, "Error: Could not open %s: %s\n", filename,
				strerror(errno));
        }
        return NULL;
    }
	
    /* Run through possibilities */
    for(/**/; *pathlist != NULL; pathlist++) {
        /* Construct absolute path name */
        if((*pathlist)[0] == '/') {
            /* This element defines an absolute path */
            snprintf(fileout, filemax, *pathlist, filename);
        } else {
		/* Not an absolute path name, append it to end of current
			* directory  */
            eol = getcwd(fileout, filemax);
            if(eol == NULL) {
                pathlist++;
                continue;
            }
            fileout[filemax - 1] = 0;
            eol += strlen(fileout);
            *(eol++) = '/';
            snprintf(eol, filemax - strlen(fileout) - 1, *pathlist, filename);
        }
        fileout[filemax - 1] = 0;
		
        /* Test for existance and access by attempting to open it */
        file = fopen(fileout, "r");
        if(file != NULL) {
            return file;
        }
        /* If failure other than file not found, report the error and quit */
        if(errno != ENOENT) {
            fprintf(stderr, "Error: Could not open %s: %s\n", filename,
				strerror(errno));
            return NULL;
        }
    }
	
    /* Could not find the file */
    fileout[0] = 0;
    fprintf(stderr, "Error: Could not find %s.\n", filename);
    return NULL;
}

inline static void printnodeinfo(xmlNodePtr node, FILE *fout) {
    assert(node != NULL);
    assert(fout != NULL);
    fprintf(fout, "%s:%ld: ", node->doc->URL, xmlGetLineNo(node));
}

static int stringtoint(const char* string, int32_t* out,
					   xmlNodePtr node) {
    char *parse;
    *out = strtol(string, &parse, 0);
    if((parse != NULL) && (*parse != 0)) {
        if(node != NULL) {
            printnodeinfo(node, stderr);
            fprintf(stderr, "error : \"%s\" is not a valid number.\n", string);
        }
        return -1;
    } else {
        return 0;
    }
}

/********************************************************************
* Functions related to loading run files
\*/

error_t atmega_parse_runfile(atmega_chip ***chips, int *nchips, char *runfile) {
    xmlDocPtr doc;
    xmlNodePtr root, curr, child, param;
    static int firstMote=-1;
    int motes, mote, result;
    xmlChar *model = NULL;
    xmlChar *flash = NULL;
    xmlChar *key = NULL;
    xmlChar *value = NULL;
    xmlChar *xmlid = NULL;
    xmlChar *title = NULL;
    ioman_param_func* paramfunc;
    device_obj* thisdev;
	
    /* Attempt to open the run file */
	//    printf(fprintf(stderr,"Runfile: %s\n", runfile);
    xmlLineNumbersDefault(1);
    doc = xmlParseFile(runfile);
    if(doc == NULL) {
        /* xmlParseFile() will have spilled the messages, so just quit */
        return -1;
    }
	
    /* Make sure it is a valid motelist */
    root = xmlDocGetRootElement(doc);
    if(root == NULL) {
        /* This shouldn't really happen... */
        return -1;
    }
    if(!IsElement(root, namespace, "motelist")) {
        printnodeinfo(root, stderr);
        fprintf(stderr, "error : Root element should be \"%s\" of"
			" namespace \"%s\", not \"%s\".\n",
			"motelist", namespace, root->name);
        xmlFreeDoc(doc);
        return -1;
    }
	
    /* Count up the number of motes */
    motes = 0;
    for(curr = root->children; curr != NULL; curr = curr->next) {
        if(IsElement(curr, namespace, "mote")) {
            motes++;
        }
    }
    if(motes < 1) {
        fprintf(stderr, "%s:0: error : File does not contain any chips.\n",
			doc->URL);
        xmlFreeDoc(doc);
        return -1;
    }
	
    /* Allocate memory for the chips */
	
    *chips = malloc(sizeof(atmega_chip*) * motes);
    g_chips=*chips;
    if(*chips == NULL){
        fprintf(stderr, "fatal error : Memory allocation failed: %s.\n",
			strerror(errno));
        xmlFreeDoc(doc);
        return -1;
    }
    *nchips = motes;
    g_nChips= motes;
    memset(*chips, 0, sizeof(atmega_chip*) * motes);
    fprintf(stderr,"Total %d sensor motes\n", *nchips);
	
    /* Load the chips */
    mote = 0;
    for(curr = root->children; curr != NULL; curr = curr->next) {
        if(IsElement(curr, namespace, "mote")) {
            /* Make sure we haven't created too many motes */
			
			
            /* Get the model filename */
            model = xmlGetProp(curr, (xmlChar*)"model");
            if(model == NULL) {
                printnodeinfo(curr, stderr);
                fprintf(stderr, "error : %s missing \"%s\" attribute.\n",
					"mote", "model");
                goto cleanup;
            }
            /* Get the flash filename */
            flash = NULL;
            for(child = curr->children; child != NULL; child = child->next) {
                if(IsElement(child, namespace, "flash")) {
                    flash = xmlGetProp(child, (xmlChar*)"href");
                    if(flash == NULL) {
                        printnodeinfo(child, stderr);
                        fprintf(stderr, "error : %s missing \"%s\""
							" attribute.\n", "flash", "href");
                        goto cleanup;
                    }
                    break;
                }
            }
			
            /* See if we have a user-specified id */ 
			// VMNet: Use the ID as the local address+++++
            xmlid = xmlGetProp(curr, (xmlChar*)"id");
            if(xmlid != NULL) {
//                fprintf(stderr,"Loading ELF for sensor mote \"%s\" (%d of %d)\n", xmlid,
//					atoi(xmlid), motes);
            } else {
                fprintf(stderr,"ERROR: mote id is incorrect! \n");
                exit(-1);
            }
            /* See if we have a user title */

            //mote=atoi(xmlid);
            //assert(mote < motes);
            if (firstMote<0)
             {
                firstMote=0;
                g_idChip=0;
              }
                else
                  g_idChip++;

            if(mote<0)
             return -2;
			//++++++++++++++++++++++++++++++++++++
            title = xmlGetProp(curr, (xmlChar*)"title");
            if(title != NULL) {
                fprintf(stderr,"\tTitle: %s\n", title);
                xmlFree(title);
            }
			fprintf(stderr,"Loading ELF for sensor mote \"%d\" (%d of %d)\n", g_idChip,
                        g_idChip, motes);
            /* Attempt to instance this particular mote */
            result = atmega_load_chip(*chips, g_idChip, (char*)model, (char*)flash);
            if(result < 0) {
          			printf("ERROR ID = %d\n",mote);
                /* Failure on the creation */
                goto cleanup;
            }
#ifdef VMNET
			// mote=0;
#endif
            (*chips)[g_idChip]->xmlid = (char*)xmlid;

            /* Loop through the devices, feeding parameters */
            for(thisdev = (*chips)[g_idChip]->device_head; thisdev != NULL;
			thisdev = thisdev->next) {
                /* Get the setparam() function */
                paramfunc = (ioman_param_func*)dlsym(thisdev->dlhandle,
					"set_param");
                if(paramfunc == NULL) {
                    continue;
                }

                /* Send in the root parameters first */
                for(param = root->children; param != NULL;
				param = param->next) {
                    if(IsElement(param, namespace, "param")) {
                        key = xmlGetProp(param, (xmlChar*)"name");
                        value = xmlGetProp(param, (xmlChar*)"value");
                        if((key != NULL) && (value != NULL)) {
                            (*paramfunc)(thisdev->dev, (char*)key,
								(char*)value);
                        }
                        if(key != NULL) {
                            xmlFree(key);
                        }
                        if(value != NULL) {
                            xmlFree(value);
                        }
                    }
                }

                /* Send in the mote-specific parameters now (overwriting
				* the common parameters) */
                for(param = curr->children; param != NULL;
				param = param->next) {
                    if(IsElement(param, namespace, "param")) {
                        key = xmlGetProp(param, (xmlChar*)"name");
                        value = xmlGetProp(param, (xmlChar*)"value");
                        if((key != NULL) && (value != NULL)) {
                            (*paramfunc)(thisdev->dev, (char*)key,
								(char*)value);
                        }
                        if(key != NULL) {
                            xmlFree(key);
                        }
                        if(value != NULL) {
                            xmlFree(value);
                        }
                    }
                }
            } /* for(thisdev = (*chips)[mote]->device_head; thisdev != NULL; */

            /* Next mote */
            xmlFree(flash);
            xmlFree(model);
			
            mote++;
        }
    }

    /* Done making motes */
    return 0;

    /* Generic clean up */
cleanup:
    /* Free allocated memory */
    xmlFreeDoc(doc);
    if(model != NULL) {
        xmlFree(model);
    }
    if(flash != NULL) {
        xmlFree(flash);
    }

    return -1;
}

/********************************************************************
* Functions related to loading configuration files
\*/

static error_t io_reg_table_lookup(char *reg, uint8_t *regnum){
    /* look up cpu code in hash table, return register number */
    io_reg *result;
    io_reg query;

    query.name=reg;

    result=(io_reg*)table_find(io_reg_table, &query);

    if(result==NULL){
        fprintf(stderr, "\tWarning: Register %s not found.\n", reg);
        return -1;
    }

    /* calculate offset into the table, that is register number */
    *regnum=result->reg;

    return 0;
}

static void regfree(void *r){
    xmlFree(((io_reg*)r)->name);
    free(r);
}

static int regcompare(const void *a, const void *b){
    return strcasecmp(((io_reg*)a)->name, ((io_reg*)b)->name);
}

static void io_reg_table_destroy(){
    table_destroy(io_reg_table, regfree);
}

static error_t set_interrupt(interrupt_obj *interrupt) {
    SET_FLAG(interrupt->chip->interrupts, interrupt->vector_mask);
    return 0;
}

static error_t clear_interrupt(interrupt_obj *interrupt) {
    CLR_FLAG(interrupt->chip->interrupts, interrupt->vector_mask);
    return 0;
}

static error_t device_add_clock(device_obj *device, clock_obj *clock) {
    /* Make sure not already in list */
    clock_obj *head;
    for(head = device->chip->clock_head; head != NULL; head = head->next) {
        if(head == clock) {
            return 0;
        }
    }

    clock->next = NULL;
    ADD_LIST(clock, device->chip->clock_head);
    return 0;
}

static error_t device_remove_clock(device_obj *device, clock_obj *clock) {
    /* Look for it in the list */
    clock_obj *head, *prev = NULL;
    for(head = device->chip->clock_head; head != NULL; head = head->next) {
        if(head == clock) {
            /* Found it */
            if(prev == NULL) {
                device->chip->clock_head = head->next;
            } else {
                prev->next = head->next;
            }
            head->next = NULL;
            return 0;
        }
        prev = head;
    }
    return 0;
}

static error_t load_ioman_device(char *soname, device_funcs *df){
    char sopath[1024];
    int i = 0;

    bzero(df, sizeof(*df));

    /* load the library */
    while((df->dlhandle == NULL) && (so_paths[i] != NULL)){
        snprintf(sopath, sizeof(sopath), so_paths[i], soname);
        df->dlhandle = dlopen(sopath, RTLD_LAZY);
        i++;
    }
    if(df->dlhandle == NULL){
        fprintf(stderr,
            "\tError: Could not find device library '%s' in search path.\n",
            /*soname*/sopath);
			return -1;
    }

    df->create_device = (ioman_create_func*)dlsym(df->dlhandle,
		"create_device");
    df->add_register = (ioman_register_func*)dlsym(df->dlhandle,
		"add_register");
    df->add_interrupt = (ioman_interrupt_func*)dlsym(df->dlhandle,
		"add_interrupt");
    df->add_clock = (ioman_clock_func*)dlsym(df->dlhandle,
		"add_clock");
    df->set_param = (ioman_param_func*)dlsym(df->dlhandle,
		"set_param");
    df->set_device = (ioman_uses_func*)dlsym(df->dlhandle,
		"set_device");

    return 0;
}

error_t atmega_parse_processor(atmega_chip *chip, xmlNodePtr proc) {
    xmlChar* name;
    xmlChar* value;
    int32_t  decode, bits;
    xmlNodePtr param;
    /* Set to true when set */
    int haveIOOffset = 0;
    int haveIOLast = 0;
    int haveSRAMBytes = 0;
    int haveFlashWords = 0;
    int havePCBits = 0;
    int havePCMask = 0;
    /* Set to true if all values set */
    int haveAll = 1;

    /* Loop through the processor children, looking for parameters */
    for(param = proc->children; param != NULL; param = param->next) {
        if(IsElement(param, namespace, "param")) {
            /* Get the parameter name */
            name = xmlGetProp(param, (xmlChar*)"name");
            if(name == NULL) {
                printnodeinfo(param, stderr);
                fprintf(stderr, "warning : %s missing \"%s\" attribute.\n",
					"param", "name");
                continue;
            }
            /* Get the parameter value (delay on decode in case a parameter
			* does not need a numerical decode) */
            value = xmlGetProp(param, (xmlChar*)"value");
            if(value == NULL) {
                xmlFree(name);
                printnodeinfo(param, stderr);
                fprintf(stderr, "warning : %s missing \"%s\" attribute.\n",
					"param", "value");
                continue;
            }

            /* Switch on parameter */
            if(strcasecmp((char*)name, "IO_OFFSET") == 0) {
                if(stringtoint((char*)value, &decode, param)) {
                    goto error;
                }
                chip->io_offset = decode;
                haveIOOffset = 1;
            } else if(strcasecmp((char*)name, "IO_LAST") == 0) {
                if(stringtoint((char*)value, &decode, param)) {
                    goto error;
                }
                chip->io_last = decode;
                haveIOLast = 1;
            } else if(strcasecmp((char*)name, "SRAM_BYTES") == 0) {
                if(stringtoint((char*)value, &decode, param)) {
                    goto error;
                }
                chip->sram_bytes = decode;
                haveSRAMBytes = 1;
            } else if(strcasecmp((char*)name, "FLASH_WORDS") == 0) {
                if(stringtoint((char*)value, &decode, param)) {
                    goto error;
                }
                chip->flash_words = decode;
                haveFlashWords = 1;
            } else if(strcasecmp((char*)name, "PC_BITS") == 0) {
                if(stringtoint((char*)value, &decode, param)) {
                    goto error;
                }
                chip->pc_bits = decode;
                havePCBits = 1;
            } else if(strcasecmp((char*)name, "PC_MASK") == 0) {
                if(stringtoint((char*)value, &decode, param)) {
                    goto error;
                }
                chip->pc_mask = decode;
                havePCMask = 1;
            }

            xmlFree(name);
            xmlFree(value);
        } /* if(IsElement(param, namespace, "param")) { */
    } /* for(param = proc->children; param != NULL; param = param->next) { */

    /* Make sure all necessary parameters have been set */
    if(!haveIOOffset) {
        printnodeinfo(proc, stderr);
        fprintf(stderr, "error : %s not set on the processor.\n",
			"IO_OFFSET");
        haveAll = 0;
    }
    if(!haveIOLast) {
        printnodeinfo(proc, stderr);
        fprintf(stderr, "error : %s not set on the processor.\n",
			"IO_LAST");
        haveAll = 0;
    }
    if(!haveSRAMBytes) {
        printnodeinfo(proc, stderr);
        fprintf(stderr, "error : %s not set on the processor.\n",
			"SRAM_BYTES");
        haveAll = 0;
    }
    if(!havePCBits) {
        if(havePCMask) {
            /* Count the number of bits used in PC_MASK */
            decode = chip->pc_mask;
            bits = 0;
            while(decode) {
                bits++;
                decode >>= 1;
            }
            havePCBits = 1;
        } else {
            printnodeinfo(proc, stderr);
            fprintf(stderr, "error : %s not set on the processor.\n",
				"PC_BITS");
            haveAll = 0;
        }
    }
    if(!havePCMask) {
        if(havePCBits) {
            /* Can be guessed from PC_BITS */
            chip->pc_mask = (1 << (chip->pc_bits - 1)) - 1;
            havePCMask = 1;
        } else {
            printnodeinfo(proc, stderr);
            fprintf(stderr, "error : %s not set on the processor.\n",
				"PC_MASK");
            haveAll = 0;
        }
    }
    if(!haveFlashWords) {
        if(havePCBits) {
            /* Can be guessed from PC_BITS */
            chip->flash_words = (1 << (chip->pc_bits - 1));
            haveFlashWords = 1;
        } else {
            printnodeinfo(proc, stderr);
            fprintf(stderr, "error : %s not set on the processor.\n",
				"FLASH_WORDS");
            haveAll = 0;
        }
    }

    /* Return an error if any parameter is missing */
    return (haveAll ? 0 : -1);

    /* Error handler */
error:
    xmlFree(name);
    xmlFree(value);
    return -1;
}

error_t atmega_parse_register(atmega_chip *chip, xmlNodePtr reg) {
    xmlChar* id;
    xmlChar* value;
    int32_t  addr;
    io_reg*  data;

    /* Get Register name */
    id = xmlGetProp(reg, (xmlChar*)"id");
    if(id == NULL) {
        printnodeinfo(reg, stderr);
        fprintf(stderr, "error : %s missing \"%s\" attribute.\n",
			"register", "id");
        return -1;
    }

    /* Get Register address */
    value = xmlGetProp(reg, (xmlChar*)"value");
    if(value == NULL) {
        printnodeinfo(reg, stderr);
        fprintf(stderr, "error : %s missing \"%s\" attribute.\n",
			"register", "value");
        xmlFree(id);
        return -1;
    }
    if(stringtoint((char*)value, &addr, reg) < 0) {
        xmlFree(value);
        xmlFree(id);
        return -1;
    }

    /* Make sure the address is in range */
    if((addr < 0) || (addr >= PARSER_MAX_REGS)){
        printnodeinfo(reg, stderr);
        fprintf(stderr, "error : Value %s out of range for register %s.\n",
			value, id);
        xmlFree(value);
        xmlFree(id);
        return -1;
    }
    xmlFree(value);

    /* Actually store the register */
    data = malloc(sizeof(*data));
    data->name = (char*)id;
    data->reg = (uint8_t)addr;
    table_store(io_reg_table, data);
    return 0;
}

error_t atmega_parse_device(atmega_chip *chip, xmlNodePtr device) {
    xmlChar* name;
    xmlChar* module;
    xmlChar* value;
    error_t err;
    xmlNodePtr param;
    device_funcs df;
    uint32_t parse_reg, parse_bits, parse_int;
    uint8_t lookup_reg;
    device_obj *dev, *search;
    register_obj *reg;
    interrupt_obj *interrupt;

    /* Get the identifier and module from the object */
    name = xmlGetProp(device, (xmlChar*)"id");
    if(name == NULL) {
        printnodeinfo(device, stderr);
        fprintf(stderr, "error : %s missing \"%s\" attribute.\n",
			"device", "id");
        return -1;
    }
    module = xmlGetProp(device, (xmlChar*)"module");
    if(module == NULL) {
        printnodeinfo(device, stderr);
        fprintf(stderr, "error : %s missing \"%s\" attribute.\n",
			"device", "module");
        xmlFree(name);
        return -1;
    }

    /* Attempt to create the object */
    err = load_ioman_device((char*)module, &df);
    if(err < 0) {
        xmlFree(name);
        xmlFree(module);
        return err;
    }

    dev = (*df.create_device)((char*)name);
    if(dev == NULL) {
        printnodeinfo(device, stderr);
        fprintf(stderr, "error : Creation of device %s failed.\n", name);
        xmlFree(name);
        xmlFree(module);
        return -1;
    }

    /* Initialize the structure */
    dev->dlhandle = df.dlhandle;
    dev->name = (char*)name;
    dev->type = (char*)module;
    dev->chip = chip;
    dev->device_event = ioman_device_event;
    dev->device_debug = ioman_device_debug;
    dev->add_clock = device_add_clock;
    dev->remove_clock = device_remove_clock;
    ADD_LIST(dev, chip->device_head);

    /* Loop through the children, looking for parameters, registers, etc. */
    for(param = device->children; param != NULL; param = param->next) {
        if(IsElement(param, NULL, NULL)) {
            if(IsElement(param, namespace, "param")) {
                /* Make sure we have a set_param */
                if(df.set_param == NULL) {
                    continue;
                }

                /* Get the parameter name */
                name = xmlGetProp(param, (xmlChar*)"name");
                if(name == NULL) {
                    printnodeinfo(device, stderr);
                    fprintf(stderr, "warning : %s missing \"%s\" attribute.\n",
						"param", "name");
                    continue;
                }

                /* Get the parameter value */
                value = xmlGetProp(param, (xmlChar*)"value");
                if(value == NULL) {
                    printnodeinfo(device, stderr);
                    fprintf(stderr, "warning : %s missing \"%s\" attribute.\n",
						"param", "value");
                    xmlFree(name);
                    continue;
                }

                /* Set the parameter */
                (*df.set_param)(dev->dev, (char*)name, (char*)value);
                xmlFree(name);
                xmlFree(value);
            } else if(IsElement(param, namespace, "register")) {
                /* Make sure we have an add_register */
                if(df.add_register == NULL) {
                    printnodeinfo(device, stderr);
                    fprintf(stderr, "error : <%s> element but"
						"%s is unimplemented in device '%s'.\n",
						"register", "add_register", dev->name);
                    return -1;
                }

                /* Get the register name */
                name = xmlGetProp(param, (xmlChar*)"name");
                if(name == NULL) {
                    printnodeinfo(device, stderr);
                    fprintf(stderr, "error : %s missing \"%s\" attribute.\n",
						"register", "name");
                    return -1;
                }

                /* Get the register port number */
                value = xmlGetProp(param, (xmlChar*)"value");
                if(value == NULL) {
                    /* Look for a register reference */
                    value = xmlGetProp(param, (xmlChar*)"ref");
                    if(value == NULL) {
                        xmlFree(name);
                        printnodeinfo(device, stderr);
                        fprintf(stderr, "error : %s requires a \"%s\" or \"%s\""
							" attributes.\n", "register", "value", "ref");
                        return -1;
                    }
                    /* Extract register reference from the table */
                    err = io_reg_table_lookup((char*)value, &lookup_reg);
                    if(err < 0) {
                        printnodeinfo(device, stderr);
                        fprintf(stderr, "error : Unknown CPU register"
							" \"%s\".\n", value);
                        xmlFree(name);
                        xmlFree(value);
                        return -1;
                    }
                    parse_reg = lookup_reg;
                } else {
                    /* Parse as if a numeric */
                    if(stringtoint((char*)value, (int32_t*)&parse_reg,
						param) < 0) {
                        xmlFree(name);
                        xmlFree(value);
                        return -1;
                    }
                }
                xmlFree(value);
                /* Make sure it is in range */
                if(parse_reg > 0xFF) {
                    printnodeinfo(device, stderr);
                    fprintf(stderr, "error : %s for register '%s' is out"
						" of bounds.\n", "Value", name);
                    xmlFree(name);
                    return -1;
                }

                /* Get the register mask */
                value = xmlGetProp(param, (xmlChar*)"bits");
                if(value == NULL) {
                    parse_bits = 0xFF;
                } else {
                    /* Parse as a numeric */
                    if(stringtoint((char*)value, (int32_t*)&parse_bits,
						param) < 0) {

                        xmlFree(name);
                        xmlFree(value);
                        return -1;
                    }
                    xmlFree(value);
                }
                /* Make sure it is in range */
                if(parse_bits > 0xFF) {
                    printnodeinfo(device, stderr);
                    fprintf(stderr, "error : %s for register '%s' is out"
						" of bounds.\n", "Bitmask", name);
                    xmlFree(name);
                    return -1;
                }
                if(parse_bits == 0) {
                    printnodeinfo(device, stderr);
                    fprintf(stderr, "warning : Bitmask for register '%s' is"
						" zero.\n", name);
                }

                /* Attempt to create register */
                reg = (*df.add_register)(dev->dev, (char*)name);
                if(reg == NULL) {
                    printnodeinfo(device, stderr);
                    fprintf(stderr, "error : Unable to add %s '%s'.\n",
						"register", name);
                    xmlFree(name);
                    return -1;
                }
                xmlFree(name);
                ADD_LIST(reg, chip->register_head[parse_reg]);

                /* Set up the mask */
                reg->mask = parse_bits;
                reg->maskshift = 0;
                while(!(parse_bits & 0x01)) {
                    reg->maskshift++;
                    parse_bits >>= 1;
                }
            } else if(IsElement(param, namespace, "interrupt")) {
                /* Make sure we have an add_interrupt */
                if(df.add_interrupt == NULL) {
                    printnodeinfo(device, stderr);
                    fprintf(stderr, "error : <%s> element but"
						"%s is unimplemented in device '%s'.\n",
						"interrupt", "add_interrupt", dev->name);
                    return -1;
                }

                /* Get the interrupt name */
                name = xmlGetProp(param, (xmlChar*)"name");
                if(name == NULL) {
                    printnodeinfo(device, stderr);
                    fprintf(stderr, "error : %s missing \"%s\" attribute.\n",
						"interrupt", "name");
                    return -1;
                }

                /* Get the interrupt number */
                value = xmlGetProp(param, (xmlChar*)"value");
                if(value == NULL) {
                    xmlFree(name);
                    printnodeinfo(device, stderr);
                    fprintf(stderr, "error : %s missing \"%s\" attribute.\n",
						"interrupt", "value");
                    return -1;
                }
                /* Parse as numeric */
                if(stringtoint((char*)value, (int32_t*)&parse_int,
					param) < 0) {
                    xmlFree(name);
                    xmlFree(value);
                    return -1;
                }
                xmlFree(value);

                /* Make sure it is in range */
                if(parse_int > 63) {
                    printnodeinfo(device, stderr);
                    fprintf(stderr, "error : Vector for interrupt '%s' is"
						" out of bounds.\n", name);
                    xmlFree(name);
                    return -1;
                }

                /* Attempt to create interrupt */
                interrupt = (*df.add_interrupt)(dev->dev, (char*)name);
                if(interrupt == NULL) {
                    printnodeinfo(device, stderr);
                    fprintf(stderr, "error : Unable to add %s '%s'.\n",
						"interrupt", name);
                    xmlFree(name);
                    return -1;
                }
                xmlFree(name);

                /* Initialize and add to list */
                interrupt->vector = parse_int;
                interrupt->vector_mask = (uint64_t)1 << parse_int;
                interrupt->chip = chip;
                interrupt->set_interrupt = set_interrupt;
                interrupt->clear_interrupt = clear_interrupt;
                ADD_LIST(interrupt, chip->interrupt_head);
            } else if(IsElement(param, namespace, "connect")) {
                /* Make sure we have an set_device */
                if(df.set_device == NULL) {
                    printnodeinfo(device, stderr);
                    fprintf(stderr, "error : <%s> element but"
						"%s is unimplemented in device '%s'.\n",
						"connect", "set_device", dev->name);
                    return -1;
                }

                /* Get the connectionname */
                name = xmlGetProp(param, (xmlChar*)"name");
                if(name == NULL) {
                    printnodeinfo(device, stderr);
                    fprintf(stderr, "error : %s missing \"%s\" attribute.\n",
						"connect", "name");
                    return -1;
                }

                /* Get the referenced device id */
                value = xmlGetProp(param, (xmlChar*)"ref");
                if(value == NULL) {
                    xmlFree(name);
                    printnodeinfo(device, stderr);
                    fprintf(stderr, "error : %s missing \"%s\" attribute.\n",
						"connect", "ref");
                    return -1;
                }

                /* Find the referenced device */
                for(search = chip->device_head; search != NULL;
				search = search->next) {
                    if(strcasecmp((char*)value, search->name) == 0) {
                        break;
                    }
                }

                if(search != NULL) {
                    /* Do the nasty with the found device */
                    err = df.set_device(dev->dev, (char*)name, search);
                    xmlFree(value);
                    xmlFree(name);
                    if(err < 0) {
                        printnodeinfo(param, stderr);
                        fprintf(stderr, "error : <connect> in device %s"
							" failed.\n", dev->name);
                        return err;
                    }
                } else {
                    /* Didn't find it */
                    printnodeinfo(param, stderr);
                    fprintf(stderr, "error : Unable to locate device \"%s\""
						" referenced by <connect> element.\n", value);
                    xmlFree(name);
                    xmlFree(value);
                    return -1;
                }
            }
        } /* if(IsElement(param, NULL, NULL)) { */
    } /* for(param = device->children; param != NULL; param = param->next) { */

    /* Success */
    return 0;
}

error_t atmega_parse_configcore(atmega_chip *chip, char *config, int depth) {
    xmlDocPtr doc;
    xmlNodePtr root, curr;
    xmlChar *text;
    xmlChar *href;
    error_t err;
    FILE* file;
#ifdef MAXPATHLEN
    char path[MAXPATHLEN];
#else
    char path[256];
#endif

    /* Attempt to open the configuration file */
    xmlLineNumbersDefault(1);
	//    fprintf(stderr,"FILE:  %s",config);
    doc = xmlParseFile(config);
    if(doc == NULL) {
        /* xmlParseFile() will have displayed errors */
        return -1;
    }

    /* Make sure it is a valid configuration */
    root = xmlDocGetRootElement(doc);
    if(root == NULL) {
        /* Should never happen */
        return -1;
    }
    if(!IsElement(root, namespace, "configuration")) {
        printnodeinfo(root, stderr);
        fprintf(stderr, "error : Root element should be \"%s\" of"
			" namespace \"%s\", not \"%s\".\n",
			"configuration", namespace, root->name);
        xmlFreeDoc(doc);
        return -1;
    }

    /* Process the document tree */
    for(curr = root->children; curr != NULL; curr = curr->next) {
        /* Only process elements */
        if(IsElement(curr, NULL, NULL)) {
            if(IsElement(curr, namespace, "message")) {
                /* Display a message */
                text = xmlGetProp(curr, (xmlChar*)"text");
                if(text != NULL) {
                    //fprintf(stderr,"\t%s\n", text);
                    xmlFree(text);
                    text = NULL;
                } else {
                    printnodeinfo(curr, stderr);
                    fprintf(stderr, "warning : %s missing \"%s\" attribute.\n",
						"message", "text");
                }
            } else if(IsElement(curr, namespace, "processor")) {
                /* Processor description */
                err = atmega_parse_processor(chip, curr);
                if(err < 0) {
                    xmlFreeDoc(doc);
                    return err;
                }
            } else if(IsElement(curr, namespace, "register")) {
                /* Register mapping */
                err = atmega_parse_register(chip, curr);
                if(err < 0) {
                    xmlFreeDoc(doc);
                    return err;
                }
            } else if(IsElement(curr, namespace, "device")) {
                /* Device description */
                err = atmega_parse_device(chip, curr);
                if(err < 0) {
                    xmlFreeDoc(doc);
                    return err;
                }
            } else if(IsElement(curr, namespace, "include")) {
                /* Include another configuration file into this one */
                href = xmlGetProp(curr, (xmlChar*)"href");
                if(href != NULL) {
                    /* Construct absolute path */
                    file = atmega_findfile((char*)href, path, sizeof(path),
						TYPE_CONFFILE);
                    xmlFree(href);
					
                    href = NULL;
                    if(file == NULL) {
                        xmlFreeDoc(doc);
                        return -1;
                    }
                    fclose(file);
					
                    /* Parse the new configuration */
                    err = atmega_parse_configcore(chip, path, depth + 1);
                    href = NULL;
                    if(err < 0) {
                        xmlFreeDoc(doc);
                        return err;
                    }
                } else {
                    printnodeinfo(curr, stderr);
                    fprintf(stderr, "error : %s missing \"%s\" attribute.\n",
						"include", "href");
                    xmlFreeDoc(doc);
                    return -1;
                }
            } else {
                /* Unknown element */
                printnodeinfo(curr, stderr);
                fprintf(stderr, "warning : Unknown element <%s>\n",
					curr->name);
            }
        } /* if(IsElement(curr, NULL, NULL)) { */
    } /* for(curr = root->children; curr != NULL; curr = curr->next) { */
	
    /* Done */
    xmlFreeDoc(doc);
    return 0;
}

error_t atmega_parse_config(atmega_chip *chip, char *config) {
    error_t err;
	
    /* Initialize structures */
    bzero(chip->register_head, sizeof(chip->register_head));
    chip->interrupt_head = NULL;
    chip->device_head = NULL;
    chip->clock_head = NULL;

    io_reg_table = table_create(regcompare);

    /* Enter into the actual configuration parser */
    err = atmega_parse_configcore(chip, config, 1);
    if(err < 0) {
        io_reg_table_destroy();
        return err;
    }
	
    /* and then do some lookups to map magic registers */
    io_reg_table_lookup("SPL", &chip->spl_io);
    io_reg_table_lookup("SPH", &chip->sph_io);
    io_reg_table_lookup("SREG", &chip->sreg_io);
    io_reg_table_lookup("RAMPZ", &chip->rampz_io);
	//VMNet: No EIND there.+++++++++++++++++++++++++++
	//    io_reg_table_lookup("EIND", &chip->eind_io);
	//++++++++++++++++++++++++++++++++++++++++++++++++
	
    /* Successful completion */
    io_reg_table_destroy();
    return 0;
}
