/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:    
*    
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

/*
* Modifications:
*
* Copyright (c) 2004-2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*   * Redistributions of source code must retain the above copyright notice,
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its
*     contributors may be used to endorse or promote products derived from
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
*/


/*

  * Modification author: Hejun Wu (vmnet@cs.ust.hk)
  
********************************************************************************/

#include <assert.h>
#include <math.h>
#include "radio.h"
/*
* This module handles a basic transmission model for the atemu's MICA2
* radio system.  Each node will have a pointer to the air, and interface
* other devices by passing this structure to these functions.
*
* Limitations:
* - Behavior at a range of '0' is...well...physically impossible
* - Antenna length is not taken into effect.
* - Radio will only transmit one frequency at a time.
* - Only possible modulation scheme is FSK
*
* Assumptions:
* - The transmission power is in dBm
* - The received power decreases with the square of distance
*
* Best practices:
* - Use W or mW for power
* - Use meters for positions
*/
//#define SEIDEL_MODEL 
#ifdef SEIDEL_MODEL
/*
d   =input('Please enter T-R separation (km): ');
d0  =input('Please enter close-in reference distance (km): ');
fc  =input('Please enter carrier frequency (MHz):');
n   =input('Please enter path loss exponent: ');
N   =input('Please enter number of desired predicted samples: ');
sigma   =input('Please enter standard deviation of the log-normal shadowing: ');
*/
#include <stdlib.h>
#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif
#define D0_REFERENCE 1
#define FC_FREQUENCY 915
#define n_EXP 4
#define N_ZERO_MEAN 1
#define SIGMA 4.03


double SeidelPathLoss(double d, double d0, double fc, double n, int N, double sigma)
{
	double lambda=300/fc;
	double PL_d0=-10*log10(lambda*lambda/((4 * M_PI*d0)*(4 * M_PI*d0))); // path loss at d0 (dB)
	double PL_d;
	double X_sigma;
	static char first = 1;
	int r ;
	if(first)
	{
		srand(100); // seed of the rand function
		first = 0;
	}
	r = rand();
	if(r%N ==0)
		r =1;
	X_sigma=sigma*(r%N); //zero-mean log-normally distributed random
	PL_d=PL_d0+10*n*log10(d/d0)+X_sigma; //path loss
	return 10*exp(-PL_d / 10);
}
#endif

/* This increments every time someone messes with the air waves */
static int32_t currentcycle = 0;
/* All the air devices are stored in a linked list */
static radio_air *firstdevice = NULL;
static radio_air *lastdevice = NULL;
/* Number of devices in the air */
static int devicecount = 0;
/* Seed for random number generator */
uint32_t ctx = 0;
#define RADIO_RANDMAX 0x7fffffff

/* Cached value of (c/(4*pi))^2 */
#define LIGHT4PI (299792458 / (4 * M_PI))
static const double lightterm = 0.004;//0.05*0.09*1*1;  // VMNet changed to general moderl
#undef LIGHT4PI

									  /*
									  * Adaption of the random number generator from FreeBSD which comes
									  * directly from "Random number generators: good ones are hard to find",
									  * Park and Miller, Communications of the ACM, vol. 31, no. 10,
									  * October 1988, p. 1195.  This is needed to ensure deterministic behavior
									  * of the radio for debugging purposes.
*/

double air_random() {
    long hi, lo, x;
	
    if(ctx == 0) {
        ctx = 123459876;
    }
	
    hi = ctx / 127773;
    lo = ctx % 127773;
    x = 16807 * lo - 2836 * hi;
    if(x < 0) {
        x += 0x7fffffff;
    }
    ctx = x;
	
    return (x % ((uint32_t)RADIO_RANDMAX + 1)) / (double)RADIO_RANDMAX;
}

double air_gaussian_random() {
    return sqrt(-2 * log(air_random())) * cos(2 * M_PI * air_random());
}

/********************************
** Internal Cache Maintenance **
********************************/

static void air_update_cache(radio_aircache *cache) {
	LOG_CURRENT_FUNCTION("radioair_air_update_cache",1);
    cache->frequency = cache->remote->frequency;
    if(cache->remote->transmitting==1) {
		LOG_CURRENT_FUNCTION("radioair_air_update_cache",2);
        /* Calculate actual reception strength */
#if 1
        cache->strength = cache->remote->strength * cache->pathloss;
		// whj remove here                / cache->frequency / cache->frequency;
#else
        cache->strength = cache->remote->strength;
#endif
		
        /* Make sure there is no net gain in power */
        if(cache->strength > cache->remote->strength) {
            cache->strength = cache->remote->strength;
        }
		cache->transmitting = 2;
    } else {
        cache->strength = 0;
		cache->transmitting = 0;
    }
    cache->lastchange = currentcycle;
}

static void air_initialize_cache(radio_air* current, radio_aircache *cache) {
    double distance;
	
    /* Calculate distance between nodes */
    LOG_CURRENT_FUNCTION("radioair_air_initialize_cache",1);
	
    distance
        = pow(current->x - cache->remote->x, 2)
        + pow(current->y - cache->remote->y, 2)
        + pow(current->z - cache->remote->z, 2);
    if(distance == 0) {
        /* Will trigger distance too short up to 1THz */
        cache->pathloss = 1E24;
    } else {
	/* Path loss is [l/4*pi*d]^2 where l=c/f
	* We have (c/4*pi)^2 cached so this becomes
		* (c/4*pi)^2*(1/d)^4*(1/f)^2 */
		//VMNet path loss is linear to /d^4 +++++++++++++++++++++++++++++++++++++++++
#ifdef SEIDEL_MODEL
		// the experiential values such as n_EXP,N_ZERO_MEAN, SIGMA can be changed by changing the maro definition
		//currently, the values are as follows:
		//D0_REFERENCE 1
		//FC_FREQUENCY 915
		// n_EXP 4
		//N_ZERO_MEAN 1
		//SIGMA 4.03
		cache->pathloss = SeidelPathLoss(distance,D0_REFERENCE, FC_FREQUENCY,n_EXP,N_ZERO_MEAN, SIGMA);
#else
		cache->pathloss = lightterm / distance / distance;
#endif
    }
    DEBUG_OUTPUT("cache  paht loss   %f ", cache->pathloss );
    air_update_cache(cache);
}

/**************************
** Air Object Functions **
**************************/

radio_air* air_create_device() {
    radio_air* current;
    radio_aircache *cache, *prev = NULL;
    /* Create initial device */
    radio_air* newnode = (radio_air*)malloc(sizeof(radio_air));
    if(newnode == NULL) {
        return NULL;
    }

    /* Initialize with "sane" values */
    newnode->x = 0;
    newnode->y = 0;
    newnode->z = 0;
    newnode->transmitting = 0;
    newnode->lastchange = currentcycle;
    newnode->freqtrue = 1;
    newnode->freqfalse = 0;
    newnode->strengthtrue = 0.0;
    newnode->strengthfalse = 0.0;
    newnode->prev = NULL;
    newnode->next = NULL;
    newnode->cache = NULL;
    newnode->attenuate = 1;
    newnode->noiselevel = 0;
	
    /* Create initial cache */
    current = firstdevice;
    while(current) {
        cache = (radio_aircache*)malloc(sizeof(radio_aircache));
        if(cache == NULL) {
            air_destroy_device(newnode);
            return NULL;
        }
		
        /* Calculate distance term */
        cache->remote = current;
        air_initialize_cache(newnode, cache);
        cache->next = NULL;
		
        if(prev) {
            prev->next = cache;
        } else {
            newnode->cache = cache;
        }
        prev = cache;
        current = current->next;
    }
	
    /* Add to cache of other pre-existing entries */
    current = firstdevice;
    while(current) {
        cache = (radio_aircache*)malloc(sizeof(radio_aircache));
        if(cache == NULL) {
            air_destroy_device(newnode);
            return NULL;
        }
		
        cache->remote = newnode;
        air_initialize_cache(current, cache);
        cache->next = current->cache;
        current->cache = cache;
        current = current->next;
    }
	
    /* Add to the linked list */
    devicecount++;
    if(firstdevice == NULL) {
        /* First device created */
        firstdevice = lastdevice = newnode;
    } else {
        /* Devices already exist, add to end */
        newnode->prev = lastdevice;
        lastdevice->next = newnode;
        lastdevice = newnode;
    }
	
    return newnode;
}

error_t air_set_position(radio_air *node, double x, double y, double z) {
    radio_air *current;
    radio_aircache *cache;
	
    /* Quick test before wasting time */
    if((node->x == x) && (node->y == y) && (node->z == z)) {
        return 0;
    }
	
    /* Update our position and update the cycle count */
    node->x = x;
    node->y = y;
    node->z = z;
    if(node->transmitting) {
        node->lastchange = ++currentcycle;
    }
	
    /* Divisors are only recalculated on position changes */
    current = firstdevice;
    while(current) {
        if(current != node) {
            /* Find this node's cache entry for the moved node */
            cache = current->cache;
            while(cache && cache->remote != node) {
                cache = cache->next;
            }
            /* Update the new position */
            if(cache) {
                air_initialize_cache(current, cache);
            }
        }
        current = current->next;
    }
	
    /* Fix our own distances */
    cache = node->cache;
    while(cache) {
        air_initialize_cache(node, cache);
        cache = cache->next;
    }
	
    /* Return success */
    return 0;
}

error_t air_destroy_device(radio_air *node) {
    radio_aircache *prev, *next;
    radio_air *current;
    /* Release its cache */
    prev = node->cache;
    while(prev) {
        next = prev->next;
        free(prev);
        prev = next;
    }
    node->cache = NULL;
	
    /* Release other nodes' cache of this node */
    current = firstdevice;
    while(current) {
        next = current->cache;
        prev = NULL;
        while(next && (next->remote != node)) {
            prev = next;
            next = next->next;
        }
		
        if(next) {
            if(prev) {
                prev->next = next->next;
            } else {
                current->cache = next->next;
            }
            free(next);
        }
		
        /* If the current node, remove from list */
        if(current == node) {
            if(current->prev) {
                current->prev->next = current->next;
            } else {
                firstdevice = current->next;
            }
			
            if(current->next) {
                current->next->prev = current->prev;
            } else {
                lastdevice = current->prev;
            }
        }
		
        current = current->next;
    }
	
    /* Free and return */
    node->next = NULL;
    node->prev = NULL;
    free(node);
    return 0;
}

/************************
** Transmit Functions **
************************/

/* set the transmission power */
error_t air_transmit_power(radio_air *node, double power){
    if(node->strength == power){
        return 0;
    }
	
    node->strength = power * node->attenuate;
    node->lastchange = ++currentcycle;
	
    return 0;
}

/* set that we are transmitting such and such a frequency */
/////////////////////////////////////////////////////////////////////////////////
//VMNet Manchester coding need this./
#ifndef WHJ_MAN_TRANSMIT
error_t air_transmit(radio_air *node, double freq/*, char mansend*/){
	if(/*(!mansend)&&*/ node->transmitting && node->frequency == freq){
        return 0;
    }
	
#else
	error_t air_transmit(radio_air *node, double freq, char mansend){
		if((!mansend)&& node->transmitting && node->frequency == freq){
			return 0;
		}
#endif
		
		LOG_CURRENT_OPERATION(RADIO_TRANSMIT,RADIO_MASK);
		LOG_CURRENT_FUNCTION("radioair_air_transmit",1);
		node->transmitting = 1;
		node->frequency = freq;
		node->lastchange = ++currentcycle;
		//DEBUG_OUTPUT("radioair_air_transmit %f", freq);
		return 0;
	}
	///////////////////////////////////////////////////////////////////////////
	/* stop transmitting */
	error_t air_transmit_release(radio_air *node){
		if(!(node->transmitting)){
			return 0;
		}
        LOG_CURRENT_FUNCTION("radioair_air_transmit_release",1);
		node->transmitting = 0;
		node->lastchange = ++currentcycle;
		
		return 0;
	}
	
	/************************
	** Transmit Functions **
	************************/
	
	/* hint as to if there's something to receive */
	int air_receive_changed(radio_air *node) {
		return (node->lastread != currentcycle);
		
	}
	
	/* set tuner params */
	error_t air_receive_tuner(radio_air *node, double frequency, double deltafreq,
        double threshold){
		if((node->frequency == frequency) && (node->deltafreq == deltafreq)
			&& (node->threshold == threshold)) {
			return 0;
		}
		node->frequency = frequency;
		node->deltafreq = deltafreq;
		node->threshold = threshold;
		node->lastchange = ++currentcycle;
		
		return 0;
	}
	
	/* If there has been a change, update the datastructures */
	error_t air_cache_update(radio_air *node) {
		if(node->lastread != currentcycle){
			radio_aircache *cache = node->cache;
			node->lastread = currentcycle;
			
			while(cache) {
				/* Update this cache entry */
				if(cache->lastchange != cache->remote->lastchange) {
					air_update_cache(cache);
				}
				cache = cache->next;
			}
		}
		
		return 0;
	}
	
	/* runs tuning on node->frequency.  finds strength, and if frequency != NULL.
	* also fills in average frequency on the air */
	error_t air_tune_freq(radio_air *node, double tunefreq, double *frequency,
        double *strength){
		radio_aircache *cache = node->cache;
		int activity = 0;
		
		if(frequency != NULL){
			*frequency = 0;
		}
		*strength = 0;
		
		/* count up weighted frequencies and signal strength */
		/* use tuned frequency and deltafreq to drop off more distant freqs */
		while(cache) {
			if(cache->transmitting==2) {
				double scalar;
				scalar = pow(10, - fabs(cache->frequency - tunefreq) / node->deltafreq);
				scalar *= cache->strength;
				*strength += scalar;
				if(frequency != NULL) {
					*frequency += cache->frequency * scalar;
				}
				activity = 1;
			}
			cache = cache->next;
		}
		
		if((frequency != NULL) && activity/*&&(*strength!=0)*/) {
			
			*frequency /= *strength;
		}
		
		return 0;
	}
	
	/* try to tune to a frequency */
	int air_receive_sense(radio_air *node, double *frequency, double *strength){
		air_cache_update(node);
		LOG_CURRENT_FUNCTION("radioair_air_receive_sense",1);
		LOG_CURRENT_OPERATION(RADIO_RECEIVE,RADIO_MASK);
		air_tune_freq(node, node->frequency, frequency, strength);
		
		/* if we found a good signal, find the weighted average */
		
		if(*strength > node->threshold){
			return 1;
		}
		
		/* no (suitable) transmissions */
		return 0;
	}
	
	/* set true and false frequencies, and transition threshold */
	error_t air_receive_frequencies(radio_air *node, double freqfalse,
        double freqtrue, double schmittdist){
		if(node->freqtrue == freqtrue && node->freqfalse == freqfalse &&
            node->schmittdist == schmittdist){
			return 0;
		}
		LOG_CURRENT_FUNCTION("radioair_air_receive_frequencies",1);
		node->freqtrue = freqtrue;
		node->freqfalse = freqfalse;
		node->schmittdist = schmittdist;
		node->lastchange = ++currentcycle;
		return 0;
	}
	
	int air_receive(radio_air *node, int *bitout, double *strengthout){
		int bit;
		static double decide=0;
		static int bitcount=0;
		double strength, snr, error;
		double tmprdm;
		/* perform tuning on both truefreq and falsefreq */
		LOG_CURRENT_OPERATION(RADIO_RECEIVE,RADIO_MASK);
		air_cache_update(node);
		air_tune_freq(node, node->freqtrue, NULL, &(node->strengthtrue));
		air_tune_freq(node, node->freqfalse, NULL, &(node->strengthfalse));
		/* Determine initial bit */
		if(node->strengthtrue > node->strengthfalse) {
			//VMNet ERROR model++++++++++++++++++++++++++++++++++++++
			
			strength = node->strengthtrue;
			#if 0
			if(node->strengthfalse)
				snr = strength / node->strengthfalse; //whj previous -
			else
				snr=0x7fffff;
			#endif
			bit = 1;
		} else {
			
			strength = node->strengthfalse;
			#if 0
			if(node->strengthtrue)
				snr = strength / node->strengthtrue; //whj previous -
			else
				snr=0x7fffff;
			
			LOG_CURRENT_FUNCTION("radioair_air_receive",3);
			#endif
			bit = 0;
		}
		#if 0
		//if(decide==0)  
		{
			snr*=((1-node->noiselevel)/node->noiselevel);
			decide=(double)exp((-snr) / 4)/2;
			
		}
		tmprdm= air_random();
		//printf("radioair_air_receive ----SNR, %.16f, %.16f \n",decide,tmprdm);
		// if(node->noiselevel>tmprdm) 
		{
			/* Determine Signal To Noise Ratio */
			
			
			//   snr /= node->noiselevel;
			/* Calculate probability of error */
			//error = exp(-snr / 4) / 2;
			
			if(tmprdm&&((bitcount)>((double)1/decide))&&(decide)>(tmprdm))
			{
				bitcount=0;
				printf("ERR bit occured \n");
				//bit = !bit; 
				
			}
			else bitcount++;
		}
		#endif
		*bitout = bit;//whj 1-bit, previous bit.
		//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
		*strengthout = strength;
		return (strength > node->threshold);
	}
	
