/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

/*
* Modifications:
*
* Copyright (c) 2004-2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*   * Redistributions of source code must retain the above copyright notice,
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its
*     contributors may be used to endorse or promote products derived from
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
*/


/*

 * Modification author: Hejun Wu (vmnet@cs.ust.hk)
 
********************************************************************************/

#include "timer16.h"
#include "../EM/debug.h"

/* List of divisors for each prescale selection */
static const uint16_t g_prescalers[] = { 0, 1, 8, 64, 256, 1024, -1, -1 };

/* Local Static Functions */
static void timer16_destroy_device(void *device);
static error_t timer16_enter_sleep(void *device, uint8_t sleepmode);
static error_t timer16_leave_sleep(void *device);
static int timer16_clearint_compareA(void *device);
static int timer16_clearint_compareB(void *device);
static int timer16_clearint_compareC(void *device);
static int timer16_clearint_overflow(void *device);
static int timer16_clearint_input(void *device);

static uint8_t timer16_controlA_read(void *dev);
static void timer16_controlA_write(void *dev, uint8_t value);
static uint8_t timer16_controlB_read(void *dev);
static void timer16_controlB_write(void *dev, uint8_t value);
static uint8_t timer16_controlC_read(void *dev);
static void timer16_controlC_write(void *dev, uint8_t value);
static uint8_t timer16_compareA_read(void *dev);
static void timer16_compareA_write(void *dev, uint8_t value);
static uint8_t timer16_compareB_read(void *dev);
static void timer16_compareB_write(void *dev, uint8_t value);
static uint8_t timer16_compareC_read(void *dev);
static void timer16_compareC_write(void *dev, uint8_t value);
static uint8_t timer16_count_read(void *dev);
static void timer16_count_write(void *dev, uint8_t value);
static uint8_t timer16_highbyte_read(void *dev);
static void timer16_highbyte_write(void *dev, uint8_t value);
static uint8_t timer16_capture_read(void *dev);
static void timer16_capture_write(void *dev, uint8_t value);
static uint8_t timer16_intmask_read(void *dev);
static void timer16_intmask_write(void *dev, uint8_t value);
static uint8_t timer16_intemask_read(void *dev);
static void timer16_intemask_write(void *dev, uint8_t value);
static uint8_t timer16_intflag_read(void *dev);
static void timer16_intflag_write(void *dev, uint8_t value);
static uint8_t timer16_inteflag_read(void *dev);
static void timer16_inteflag_write(void *dev, uint8_t value);
static uint8_t timer16_sfior_read(void *dev);
static void timer16_sfior_write(void *dev, uint8_t value);
static int timer16_check_control(timer16_device *timer);
static int timer16_clock_cpu(void *device);
static int timer16_clock_external(void *device);
static int timer16_clock_common(timer16_device *timer);

/*
* Startup functions
*/
device_obj* create_device(char *name) {
	static sCounter=0;
    timer16_device *new_dev = (timer16_device*)malloc(sizeof(timer16_device));
    if(new_dev == NULL) {
        return NULL;
    }
	
    memset(new_dev, 0, sizeof(timer16_device));

    new_dev->device.destroy = timer16_destroy_device;
    new_dev->device.sleep = timer16_enter_sleep;
    new_dev->device.wake = timer16_leave_sleep;
    new_dev->device.dev = new_dev;
    new_dev->clk_tick.tick = NULL;
    new_dev->clk_tick.dev = new_dev;
    new_dev->int_compareA.clear = timer16_clearint_compareA;
    new_dev->int_compareA.dev = new_dev;
    new_dev->int_compareB.clear = timer16_clearint_compareB;
    new_dev->int_compareB.dev = new_dev;
    new_dev->int_compareC.clear = timer16_clearint_compareC;
    new_dev->int_compareC.dev = new_dev;
    new_dev->int_overflow.clear = timer16_clearint_overflow;
    new_dev->int_overflow.dev = new_dev;
    new_dev->int_input.clear = timer16_clearint_input;
    new_dev->int_input.dev = new_dev;
	
    new_dev->reg_controlA.read = timer16_controlA_read;
    new_dev->reg_controlA.write = timer16_controlA_write;
    new_dev->reg_controlA.dev = new_dev;
    new_dev->reg_controlB.read = timer16_controlB_read;
    new_dev->reg_controlB.write = timer16_controlB_write;
    new_dev->reg_controlB.dev = new_dev;
    new_dev->reg_controlC.read = timer16_controlC_read;
    new_dev->reg_controlC.write = timer16_controlC_write;
    new_dev->reg_controlC.dev = new_dev;
    new_dev->reg_compareA.read = timer16_compareA_read;
    new_dev->reg_compareA.write = timer16_compareA_write;
    new_dev->reg_compareA.dev = new_dev;
    new_dev->reg_compareB.read = timer16_compareB_read;
    new_dev->reg_compareB.write = timer16_compareB_write;
    new_dev->reg_compareB.dev = new_dev;
    new_dev->reg_compareC.read = timer16_compareC_read;
    new_dev->reg_compareC.write = timer16_compareC_write;
    new_dev->reg_compareC.dev = new_dev;
    new_dev->reg_count.read = timer16_count_read;
    new_dev->reg_count.write = timer16_count_write;
    new_dev->reg_count.dev = new_dev;
    new_dev->reg_capture.read = timer16_capture_read;
    new_dev->reg_capture.write = timer16_capture_write;
    new_dev->reg_capture.dev = new_dev;
    new_dev->reg_highbyte.read = timer16_highbyte_read;
    new_dev->reg_highbyte.write = timer16_highbyte_write;
    new_dev->reg_highbyte.dev = new_dev;
    new_dev->reg_intmask.read = timer16_intmask_read;
    new_dev->reg_intmask.write = timer16_intmask_write;
    new_dev->reg_intmask.dev = new_dev;
    new_dev->reg_intemask.read = timer16_intemask_read;
    new_dev->reg_intemask.write = timer16_intemask_write;
    new_dev->reg_intemask.dev = new_dev;
    new_dev->reg_intflag.read = timer16_intflag_read;
    new_dev->reg_intflag.write = timer16_intflag_write;
    new_dev->reg_intflag.dev = new_dev;
    new_dev->reg_inteflag.read = timer16_inteflag_read;
    new_dev->reg_inteflag.write = timer16_inteflag_write;
    new_dev->reg_inteflag.dev = new_dev;
    new_dev->reg_sync.read = timer16_sfior_read;
    new_dev->reg_sync.write = timer16_sfior_write;
    new_dev->reg_sync.dev = new_dev;
	
    new_dev->emaskonly = 1;
	
    /* Initialize device to sane values */
    new_dev->controlA = 0;
    new_dev->controlB = 0;
    new_dev->controlC = 0;
    new_dev->count = 0;
    new_dev->compareA = 0;
    new_dev->compareB = 0;
    new_dev->compareC = 0;
    new_dev->capture = 0;
    timer16_check_control(new_dev);
    
    return &(new_dev->device);
}

register_obj *add_register(void *dev, char *name) {
    timer16_device *timer = (timer16_device*)dev;
	
    if(strcasecmp(name, "CONTROLA") == 0) {
        return &(timer->reg_controlA);
    } else if(strcasecmp(name, "CONTROLB") == 0) {
        return &(timer->reg_controlB);
    } else if(strcasecmp(name, "CONTROLC") == 0) {
        return &(timer->reg_controlC);
    } else if(strcasecmp(name, "COMPAREA") == 0) {
        return &(timer->reg_compareA);
    } else if(strcasecmp(name, "COMPAREB") == 0) {
        return &(timer->reg_compareB);
    } else if(strcasecmp(name, "COMPAREC") == 0) {
        return &(timer->reg_compareC);
    } else if(strcasecmp(name, "COUNT") == 0) {
        return &(timer->reg_count);
    } else if(strcasecmp(name, "HIGHBYTE") == 0) {
        return &(timer->reg_highbyte);
    } else if(strcasecmp(name, "INTMASK") == 0) {
        timer->emaskonly = 0;
        return &(timer->reg_intmask);
    } else if(strcasecmp(name, "INTFLAG") == 0) {
        timer->emaskonly = 0;
        return &(timer->reg_intflag);
    } else if(strcasecmp(name, "INTEMASK") == 0) {
        return &(timer->reg_intemask);
    } else if(strcasecmp(name, "INTEFLAG") == 0) {
        return &(timer->reg_inteflag);
    } else if(strcasecmp(name, "SFIOR") == 0) {
        return &(timer->reg_sync);
    }
	
    return NULL;
}

interrupt_obj* add_interrupt(void *dev, char *name) {
    timer16_device *timer = (timer16_device*)dev;
	
    if(strcasecmp(name, "COMPAREA") == 0) {
        return &(timer->int_compareA);
    } else if(strcasecmp(name, "COMPAREB") == 0) {
        return &(timer->int_compareB);
    } else if(strcasecmp(name, "COMPAREC") == 0) {
        return &(timer->int_compareC);
    } else if(strcasecmp(name, "OVERFLOW") == 0) {
        return &(timer->int_overflow);
    } else if(strcasecmp(name, "INPUT") == 0) {
        return &(timer->int_input);
    }
	
    return NULL;
}

clock_obj* add_clock(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
    return &(timer->clk_tick);
}

/*
* Basic device callbacks
*/

static void timer16_destroy_device(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
	
    free(timer);
}

static error_t timer16_enter_sleep(void *dev, uint8_t sleepmode) {
    timer16_device *timer = (timer16_device*)dev;
	
    switch(sleepmode) {
    case SLEEPMODE_IDLE:
        timer->sleeping = 0;
		
        break;
    case SLEEPMODE_ADCNOISE:
    case SLEEPMODE_EXTSTANDBY:
    case SLEEPMODE_POWERDOWN:
    case SLEEPMODE_POWERSAVE:
    case SLEEPMODE_STANDBY:
		timer->sleeping = 1;
        break;
    default:
        /* Shouldn't be in this mode.... */
        timer->device.device_debug(&timer->device, V_NORMAL,
			"Unknown sleep mode %d!\n", sleepmode);
        timer->sleeping = 1;
        device_clrclock(timer->device, timer->clk_tick);
        return -1;
    }
	
    /* Disable clock as appropriate */
    if(timer->sleeping) {
        device_clrclock(timer->device, timer->clk_tick);
    }
	
    return 0;
}

static error_t timer16_leave_sleep(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
    timer->sleeping = 0;
	
    /* Reenable any normal modes */
	
	
    if(timer->prescaler_divisor > 0) {
        device_setclock(timer->device, timer->clk_tick, timer16_clock_cpu);
    }
    return 0;
}

static int timer16_clearint_compareA(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
    timer->intCompareA = 0;
	
    return 0;
}

static int timer16_clearint_compareB(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
    timer->intCompareB = 0;
	
    return 0;
}

static int timer16_clearint_compareC(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
    timer->intCompareC = 0;
	
    return 0;
}

static int timer16_clearint_overflow(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
    timer->intOverflow = 0;
    return 0;
}

static int timer16_clearint_input(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
    timer->intCapture = 0;
    return 0;
}

/*
* Device I/O
*/

static uint8_t timer16_count_read(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
    /* Count read always gives the current count, even if a previous write
	* is buffering a different count. */
    timer->temp = (uint8_t)(timer->count >> 8);
    return (uint8_t)(timer->count & 0x00FF);
}

static void timer16_count_write(void *dev, uint8_t value) {
    timer16_device *timer = (timer16_device*)dev;
    timer->count = ((uint16_t)value | ((uint16_t)(timer->temp) << 8));
}

static uint8_t timer16_capture_read(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
    /* Count read always gives the current capture, even if a previous write
	* is buffering a different capture. */
    timer->temp = (uint8_t)(timer->capture >> 8);
    return (uint8_t)(timer->capture & 0x00FF);
}

static void timer16_capture_write(void *dev, uint8_t value) {
    timer16_device *timer = (timer16_device*)dev;
    timer->capture = ((uint16_t)value | ((uint16_t)(timer->temp) << 8));
}

static uint8_t timer16_highbyte_read(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
    return timer->temp;
}

static void timer16_highbyte_write(void *dev, uint8_t value) {
    timer16_device *timer = (timer16_device*)dev;
	
    timer->temp = value;
}

static uint8_t timer16_compareA_read(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
    if(timer->buffercompare) {
        timer->temp = (uint8_t)(timer->compareAbuf >> 8);
        return (uint8_t)(timer->compareAbuf & 0x00FF);
    } else {
        timer->temp = (uint8_t)(timer->compareA >> 8);
        return (uint8_t)(timer->compareA & 0x00FF);
    }
}

static void timer16_compareA_write(void *dev, uint8_t value) {
    timer16_device *timer = (timer16_device*)dev;
    if(timer->buffercompare) {
        timer->compareAbuf = ((uint16_t)value | ((uint16_t)(timer->temp) << 8));
    } else {
        timer->compareA = ((uint16_t)value | ((uint16_t)(timer->temp) << 8));
        timer->skipnextcompA = 1;
    }
}

static uint8_t timer16_compareB_read(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
    if(timer->buffercompare) {
        timer->temp = (uint8_t)(timer->compareBbuf >> 8);
        return (uint8_t)(timer->compareBbuf & 0x00FF);
    } else {
        timer->temp = (uint8_t)(timer->compareB >> 8);
        return (uint8_t)(timer->compareB & 0x00FF);
    }
}

static void timer16_compareB_write(void *dev, uint8_t value) {
    timer16_device *timer = (timer16_device*)dev;
    if(timer->buffercompare) {
        timer->compareBbuf = ((uint16_t)value | ((uint16_t)(timer->temp) << 8));
    } else {
        timer->compareB = ((uint16_t)value | ((uint16_t)(timer->temp) << 8));
        timer->skipnextcompB = 1;
    }
}

static uint8_t timer16_compareC_read(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
    if(timer->buffercompare) {
        timer->temp = (uint8_t)(timer->compareCbuf >> 8);
        return (uint8_t)(timer->compareCbuf & 0x00FF);
    } else {
        timer->temp = (uint8_t)(timer->compareC >> 8);
        return (uint8_t)(timer->compareC & 0x00FF);
    }
}

static void timer16_compareC_write(void *dev, uint8_t value) {
    timer16_device *timer = (timer16_device*)dev;
    if(timer->buffercompare) {
        timer->compareCbuf = ((uint16_t)value | ((uint16_t)(timer->temp) << 8));
    } else {
        timer->compareC = ((uint16_t)value | ((uint16_t)(timer->temp) << 8));
        timer->skipnextcompC = 1;
    }
}

static uint8_t timer16_controlA_read(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
    return timer->controlA;
}

static void timer16_controlA_write(void *dev, uint8_t value) {
    timer16_device *timer = (timer16_device*)dev;
	
    /* Copy and perform basic tests */
    timer->controlA = value;
    timer16_check_control(timer);
}

static uint8_t timer16_controlB_read(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
    return timer->controlB;
    
}

static void timer16_controlB_write(void *dev, uint8_t value) {
    timer16_device *timer = (timer16_device*)dev;
	
    /* Copy and perform basic tests */
    timer->controlB = value;
    timer16_check_control(timer);
}

static uint8_t timer16_controlC_read(void *dev) {
    /* Only has Output Compare */
    return 0;
}

static void timer16_controlC_write(void *dev, uint8_t value) {
    /* Only has Output Compare */
}

static uint8_t timer16_intflag_read(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
    return (timer->intCapture ? TIMER16_MASK_TICIE1 : 0)
        | (timer->intCompareA ? TIMER16_MASK_OCIE1A : 0)
        | (timer->intCompareB ? TIMER16_MASK_OCIE1B : 0)
        | (timer->intOverflow ? TIMER16_MASK_TOIE1 : 0);
}

static void timer16_intflag_write(void *dev, uint8_t value) {
    timer16_device *timer = (timer16_device*)dev;
    if(GET_FLAG(value, TIMER16_MASK_TICIE1)) {
        timer->intCapture = 0;
    }
    if(GET_FLAG(value, TIMER16_MASK_OCIE1A)) {
        timer->intCompareA = 0;
    }
    if(GET_FLAG(value, TIMER16_MASK_OCIE1B)) {
        timer->intCompareB = 0;
    }
    if(GET_FLAG(value, TIMER16_MASK_TOIE1)) {
        timer->intOverflow = 0;
    }
}

static uint8_t timer16_inteflag_read(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
    if(timer->emaskonly) {
        return (timer->intCapture ? TIMER16_EMASK_TICIE3 : 0)
            | (timer->intCompareA ? TIMER16_EMASK_OCIE3A : 0)
            | (timer->intCompareB ? TIMER16_EMASK_OCIE3B : 0)
            | (timer->intOverflow ? TIMER16_EMASK_TOIE3 : 0)
            | (timer->intCompareC ? TIMER16_EMASK_OCIE3C : 0);
    } else {
        return (timer->intCompareC ? TIMER16_EMASK_OCIE1C : 0);
    }
}

static void timer16_inteflag_write(void *dev, uint8_t value) {
    timer16_device *timer = (timer16_device*)dev;
    if(timer->emaskonly) {
        if(GET_FLAG(value, TIMER16_EMASK_TICIE3)) {
            timer->intCapture = 0;
        }
        if(GET_FLAG(value, TIMER16_EMASK_OCIE3A)) {
            timer->intCompareA = 0;
        }
        if(GET_FLAG(value, TIMER16_EMASK_OCIE3B)) {
            timer->intCompareB = 0;
        }
        if(GET_FLAG(value, TIMER16_EMASK_TOIE3)) {
            timer->intOverflow = 0;
        }
        if(GET_FLAG(value, TIMER16_EMASK_OCIE3C)) {
            timer->intCompareC = 0;
        }
    } else {
        if(GET_FLAG(value, TIMER16_EMASK_OCIE1C)) {
            timer->intCompareC = 0;
        }
    }
}

static uint8_t timer16_intmask_read(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
    return (timer->maskCapture ? TIMER16_MASK_TICIE1 : 0)
        | (timer->maskCompareA ? TIMER16_MASK_OCIE1A : 0)
        | (timer->maskCompareB ? TIMER16_MASK_OCIE1B : 0)
        | (timer->maskOverflow ? TIMER16_MASK_TOIE1 : 0);
}

static void timer16_intmask_write(void *dev, uint8_t value) {
    timer16_device *timer = (timer16_device*)dev;
    timer->maskCapture = GET_FLAG(value, TIMER16_MASK_TICIE1);
    timer->maskCompareA = GET_FLAG(value, TIMER16_MASK_OCIE1A);
    timer->maskCompareB = GET_FLAG(value, TIMER16_MASK_OCIE1B);
    timer->maskOverflow = GET_FLAG(value, TIMER16_MASK_TOIE1);
    if(timer->maskCapture && timer->intCapture) {
        timer->int_input.set_interrupt(& timer->int_input);
    } else {
        timer->int_input.clear_interrupt(& timer->int_input);
    }
    if(timer->maskCompareA && timer->intCompareA) {
        timer->int_compareA.set_interrupt(& timer->int_compareB);
    } else {
        timer->int_compareA.clear_interrupt(& timer->int_compareA);
    }
    if(timer->maskCompareB && timer->intCompareB) {
        timer->int_compareB.set_interrupt(& timer->int_compareB);
    } else {
        timer->int_compareB.clear_interrupt(& timer->int_compareB);
    }
    if(timer->maskOverflow && timer->intOverflow) {
        timer->int_overflow.set_interrupt(& timer->int_overflow);
    } else {
        timer->int_overflow.clear_interrupt(& timer->int_overflow);
    }
}

static uint8_t timer16_intemask_read(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
    if(timer->emaskonly) {
        return (timer->maskCapture ? TIMER16_EMASK_TICIE3 : 0)
            | (timer->maskCompareA ? TIMER16_EMASK_OCIE3A : 0)
            | (timer->maskCompareB ? TIMER16_EMASK_OCIE3B : 0)
            | (timer->maskOverflow ? TIMER16_EMASK_TOIE3 : 0)
            | (timer->maskCompareC ? TIMER16_EMASK_OCIE3C : 0);
    } else {
        return (timer->maskCompareC ? TIMER16_EMASK_OCIE1C : 0);
    }
}

static void timer16_intemask_write(void *dev, uint8_t value) {
    timer16_device *timer = (timer16_device*)dev;
    if(timer->emaskonly) {
        timer->maskCapture = GET_FLAG(value, TIMER16_EMASK_TICIE3);
        timer->maskCompareA = GET_FLAG(value, TIMER16_EMASK_OCIE3A);
        timer->maskCompareB = GET_FLAG(value, TIMER16_EMASK_OCIE3B);
        timer->maskOverflow = GET_FLAG(value, TIMER16_EMASK_TOIE3);
        timer->maskCompareC = GET_FLAG(value, TIMER16_EMASK_OCIE3C);
        if(timer->maskCapture && timer->intCapture) {
            timer->int_input.set_interrupt(& timer->int_input);
        } else {
            timer->int_input.clear_interrupt(& timer->int_input);
        }
        if(timer->maskCompareA && timer->intCompareA) {
            timer->int_compareA.set_interrupt(& timer->int_compareB);
        } else {
            timer->int_compareA.clear_interrupt(& timer->int_compareA);
        }
        if(timer->maskCompareB && timer->intCompareB) {
            timer->int_compareB.set_interrupt(& timer->int_compareB);
        } else {
            timer->int_compareB.clear_interrupt(& timer->int_compareB);
        }
        if(timer->maskOverflow && timer->intOverflow) {
            timer->int_overflow.set_interrupt(& timer->int_overflow);
        } else {
            timer->int_overflow.clear_interrupt(& timer->int_overflow);
        }
    } else {
        timer->maskCompareC = GET_FLAG(value, TIMER16_EMASK_OCIE1C);
    }
	
    if(timer->maskCompareC && timer->intCompareC) {
        timer->int_compareC.set_interrupt(& timer->int_compareC);
    } else {
        timer->int_compareC.clear_interrupt(& timer->int_compareC);
    }
}

static uint8_t timer16_sfior_read(void *dev) {
    timer16_device *timer = (timer16_device*)dev;
    return ((timer->enabletsm) ? TIMER16_SFIOR_TSM : 0)
        | ((timer->enablepsr) ? TIMER16_SFIOR_PSR : 0);
}

static void timer16_sfior_write(void *dev, uint8_t value) {
    timer16_device *timer = (timer16_device*)dev;
    /* Prescaler held in reset if PSR set */
    if(value & TIMER16_SFIOR_PSR) {
        timer->enablepsr = 1;
        timer->clock_divide = 0;
    } else {
        timer->enablepsr = 0;
    }
    /* If TSM is cleared, all clocks restart */
    if(value & TIMER16_SFIOR_TSM) {
        timer->enabletsm = 1;
    } else {
        timer->enabletsm = 0;
        if(timer->enabletsm) {
            timer->enablepsr = 0;
        }
    }
}

/*
* Main hardware logic
*/
static int timer16_check_control(timer16_device *timer) {
    /* See if we are in a PWM mode (compares are buffered) */
    if(((timer->compareA & TIMER16_CONTROLA_WGM1) == 0) 
		&& ((timer->compareB & TIMER16_CONTROLB_WGM1)
		!= TIMER16_CONTROLB_WGM13)) {
        timer->buffercompare = 0;
    } else {
        timer->buffercompare = 1;
    }
    /* Select the prescaler */
    timer->prescaler_divisor
        = g_prescalers[timer->controlB & TIMER16_CONTROLB_CS1];
    /* Check if timer being disabled or disabled */
    if(timer->device.add_clock) {
        if(timer->prescaler_divisor == 0) {
            /* Clock disabled */
            device_clrclock(timer->device, timer->clk_tick);
            timer->externalclock = 0;
        } else if(timer->prescaler_divisor != (uint16_t)-1) {
            /* Clocked by CPU clock */
            
            device_setclock(timer->device, timer->clk_tick, timer16_clock_cpu);
            timer->externalclock = 0;
        } else {
            /* External clock generation */
            device_clrclock(timer->device, timer->clk_tick);
            timer->externalclock = 1;
        }
    }
	
    /* Select the appropriate TOP */
    switch(timer->controlB & TIMER16_CONTROLB_WGM1) {
    case 0:
        /* High bits 00xx */
        switch(timer->controlA & TIMER16_CONTROLA_WGM1) {
        case 0:
            /* Normal Mode: WGM1=0000 */
            timer->top = &(timer->max);
            timer->overflow = &(timer->max);
            timer->max = 0xFFFF;
            timer->mode = TIMER16_MODE_NORMAL;
            break;
        case TIMER16_CONTROLA_WGM10:
            /* PWM, Phase Correct, 8-bit: WGM1=0001 */
            timer->top = &(timer->max);
            timer->overflow = &(timer->bottom);
            timer->max = 0x00FF;
            timer->bottom = 0x0000;
            timer->mode = TIMER16_MODE_PHASEPWM;
            break;
        case TIMER16_CONTROLA_WGM11:
            /* PWM, Phase Correct, 9-bit: WGM1=0010 */
            timer->top = &(timer->max);
            timer->overflow = &(timer->bottom);
            timer->max = 0x01FF;
            timer->bottom = 0x0000;
            timer->mode = TIMER16_MODE_PHASEPWM;
            break;
        case TIMER16_CONTROLA_WGM1:
            /* PWM, Phase Correct, 10-bit: WGM1=0011 */
            timer->top = &(timer->max);
            timer->overflow = &(timer->bottom);
            timer->max = 0x03FF;
            timer->bottom = 0x0000;
            timer->mode = TIMER16_MODE_PHASEPWM;
            break;
        }
        break;
		case TIMER16_CONTROLB_WGM12:
			/* High bits 01xx */
			switch(timer->controlA & TIMER16_CONTROLA_WGM1) {
			case 0:
				/* CTC: WGM1=0100 */
				timer->top = &(timer->compareA);
				timer->overflow = timer->top;
				timer->max = 0xFFFF;
				timer->mode = TIMER16_MODE_CTC;
				break;
			case TIMER16_CONTROLA_WGM10:
				/* Fast PWM, 8-bit: WGM1=0101 */
				timer->top = &(timer->max);
				timer->overflow = timer->top;
				timer->max = 0x00FF;
				timer->mode = TIMER16_MODE_FASTPWM;
				break;
			case TIMER16_CONTROLA_WGM11:
				/* Fast PWM, 9-bit: WGM1=0110 */
				timer->top = &(timer->max);
				timer->overflow = timer->top;
				timer->max = 0x01FF;
				timer->mode = TIMER16_MODE_FASTPWM;
				break;
			case TIMER16_CONTROLA_WGM1:
				/* Fast PWM, 10-bit: WGM1=0111 */
				timer->top = &(timer->max);
				timer->overflow = timer->top;
				timer->max = 0x03FF;
				timer->mode = TIMER16_MODE_FASTPWM;
				break;
			}
			break;
			case TIMER16_CONTROLB_WGM13:
				/* High bits 10xx */
				switch(timer->controlA & TIMER16_CONTROLA_WGM1) {
				case 0:
					/* PWM, Phase and Frequency Correct: WGM1=1000 */
					timer->top = &(timer->capture);
					timer->overflow = &(timer->bottom);
					timer->bottom = 0x0000;
					timer->mode = TIMER16_MODE_FREQPWM;
					break;
				case TIMER16_CONTROLA_WGM10:
					/* PWM, Phase and Frequency Correct: WGM1=1001 */
					timer->top = &(timer->compareA);
					timer->overflow = &(timer->bottom);
					timer->bottom = 0x0000;
					timer->mode = TIMER16_MODE_FREQPWM;
					break;
				case TIMER16_CONTROLA_WGM11:
					/* PWM, Phase Correct: WGM1=1010 */
					timer->top = &(timer->capture);
					timer->overflow = &(timer->bottom);
					timer->bottom = 0x0000;
					timer->mode = TIMER16_MODE_PHASEPWM;
					break;
				case TIMER16_CONTROLA_WGM1:
					/* PWM, Phase Correct: WGM1=1011 */
					timer->top = &(timer->compareA);
					timer->overflow = &(timer->bottom);
					timer->bottom = 0x0000;
					timer->mode = TIMER16_MODE_PHASEPWM;
					break;
				}
				break;
				case TIMER16_CONTROLB_WGM1:
					/* High bits 11xx */
					switch(timer->controlA & TIMER16_CONTROLA_WGM1) {
					case 0:
						/* CTC: WGM1=1100 */
						timer->top = &(timer->capture);
						timer->overflow = &(timer->max);
						timer->max = 0xFFFF;
						timer->mode = TIMER16_MODE_CTC;
						break;
					case TIMER16_CONTROLA_WGM10:
						/* Reserved: WGM1=1101 */
						timer->top = &(timer->max);
						timer->max = 0xFFFF;
						/* Disable the clock in an illegal mode */
						device_clrclock(timer->device, timer->clk_tick);
						timer->externalclock = 0;
						timer->prescaler_divisor = 0;
						break;
					case TIMER16_CONTROLA_WGM11:
						/* Fast PWM: WGM1=1110 */
						timer->top = &(timer->capture);
						timer->overflow = timer->top;
						timer->mode = TIMER16_MODE_FASTPWM;
						break;
					case TIMER16_CONTROLA_WGM1:
						/* Fast PWM: WGM1=1111 */
						timer->top = &(timer->compareA);
						timer->overflow = timer->top;
						timer->mode = TIMER16_MODE_FASTPWM;
						break;
					}
					break;
    }
	
    return 0;
}

static int timer16_clock_cpu(void *device) {
    timer16_device *timer = (timer16_device*)device;
    uint8_t prescaler, prescalebit;
	
    /* Run through the prescaler */
    prescaler = g_prescalers[timer->controlB & TIMER16_CONTROLB_CS1];
    if(prescaler == 0) {
        /* The clock is disabled, why are we here? */
		
        device_clrclock(timer->device, timer->clk_tick);
    } else {
        /* Running through the prescaler */
        if(!timer->enablepsr) {
            /* If Prescale Reset is disabled, increment */
            timer->clock_divide++;
        } else if(!timer->enabletsm) {
            /* If PSR is set without TSM, PSR is cleared after one tick */
            timer->enablepsr = 0;
        }
        /* Use the list for timer 0 */
        prescalebit = GET_FLAG(timer->clock_divide, prescaler);
        /* The logic sees a tick occurs if the bit has toggled since the
		* last tick */
        if(prescalebit != timer->lasttick) {
            timer16_clock_common(timer);
        }
        timer->lasttick = prescalebit;
    }
	
    return 0;
}

static int timer16_clock_external(void *device) {
    timer16_device *timer = (timer16_device*)device;
	
    /* Only called on change */
    device_clrclock(timer->device, timer->clk_tick);
	
    /* Make sure there has been a change */
    if(timer->externalvalue == timer->externalprevious) {
        return 0;
    }
	
    /* Make sure it is the correct transition */
    timer->externalprevious = timer->externalvalue;
    if(timer->externalvalue) {
        /* Rising edge */
        if(GET_FLAG(timer->controlB, TIMER16_CONTROLB_CS10)) {
            /* We clock on rising edge */
            timer16_clock_common(timer);
        }
    } else {
        /* Falling edge */
        if(!GET_FLAG(timer->controlB, TIMER16_CONTROLB_CS10)) {
            /* We clock on falling edge */
            timer16_clock_common(timer);
        }
    }
	
    return 0;
}

error_t timer16_drive_clock(timer16_device* timer, pin_drive_enum value) {
    unsigned bit = GET_FLAG(value, PIN_DRIVE_HIGH);
    /* For efficiency, only perform above function on clock change */
    if(bit != timer->externalvalue) {
        timer->externalvalue = bit;
        /* Don't touch clk_tick if not in an externally clocked mode */
        if(timer->externalclock) {
            device_setclock(timer->device, timer->clk_tick,
				timer16_clock_external);
        }
    }
	
    return 0;
}

static int timer16_clock_common(timer16_device *timer) {
    uint8_t intcompA = 0, intcompB = 0, intcompC = 0, intover = 0;
	
    /* Comparision made continuously */
    intcompA = !timer->skipnextcompA && (timer->count == timer->compareA);
    intcompB = !timer->skipnextcompB && (timer->count == timer->compareB);
    intcompC = !timer->skipnextcompC && (timer->count == timer->compareC);
    /* Compares are skipped after a new count is written */
    timer->skipnextcompA = 0;
    timer->skipnextcompB = 0;
    timer->skipnextcompC = 0;
    /* Overflow comparision */
    intover = (timer->count == *(timer->overflow));
	
    /* Control Logic (overflow handling is specific to the waveform
	* generation mode, and is performed before increment.  It is supposed
	* to be timed for the tick /after/ the event actually occured) */
    switch(timer->mode) {
    case TIMER16_MODE_NORMAL:
        /* Normal Mode */
        timer->count++;
        break;
    case TIMER16_MODE_CTC:
        /* Clear Timer on Compare */
        if(timer->count == *(timer->top)) {
            timer->count = 0x0000;
        } else {
            timer->count++;
        }
        break;
    case TIMER16_MODE_FASTPWM:
        /* Fast PWM */
        if(timer->count == *(timer->top)) {
            timer->count = 0x0000;
            /* Compare is buffered until TOP */
            timer->compareA = timer->compareAbuf;
            timer->compareB = timer->compareBbuf;
            timer->compareC = timer->compareCbuf;
        } else {
            timer->count++;
        }
        break;
    case TIMER16_MODE_PHASEPWM:
        /* Phase Correct PWM */
        if(timer->count == *(timer->top)) {
            timer->direction = 1;
            /* Compare is buffered until TOP */
            timer->compareA = timer->compareAbuf;
            timer->compareB = timer->compareBbuf;
            timer->compareC = timer->compareCbuf;
        } else if(timer->count == 0x0000) {
            timer->direction = 0;
        }
        if(timer->direction) {
            timer->count--;
        } else {
            timer->count++;
        }
        break;
    case TIMER16_MODE_FREQPWM:
        /* Frequency and Phase Correct PWM */
        if(timer->count == *(timer->top)) {
            timer->direction = 1;
        } else if(timer->count == 0x0000) {
            timer->direction = 0;
            /* Compare is buffered until BOTTOM */
            timer->compareA = timer->compareAbuf;
            timer->compareB = timer->compareBbuf;
            timer->compareC = timer->compareCbuf;
        }
        if(timer->direction) {
            timer->count--;
        } else {
            timer->count++;
        }
        break;
    }
	
    /* Handle new interrupts */
    if(intover) {
        if(!timer->lastOverflow) {
            timer->lastOverflow = 1;
            timer->intOverflow = 1;
            if(timer->maskOverflow) {
                timer->int_overflow.set_interrupt(& timer->int_overflow);
            }
        }
    } else {
        timer->lastOverflow = 0;
    }
    if(intcompA) {
        if(!timer->lastCompareA) {
            timer->lastCompareA = 1;
            timer->intCompareA = 1;
            if(timer->maskCompareA) {
                timer->int_overflow.set_interrupt(& timer->int_compareA);
            }
        }
    } else {
        timer->lastCompareA = 0;
    }
    if(intcompB) {
        if(!timer->lastCompareB) {
            timer->lastCompareB = 1;
            timer->intCompareB = 1;
            if(timer->maskCompareB) {
                timer->int_overflow.set_interrupt(& timer->int_compareB);
            }
        }
    } else {
        timer->lastCompareB = 0;
    }
    if(intcompC) {
        if(!timer->lastCompareC) {
            timer->lastCompareC = 1;
            timer->intCompareC = 1;
            if(timer->maskCompareC) {
                timer->int_overflow.set_interrupt(& timer->int_compareC);
            }
        }
    } else {
        timer->lastCompareC = 0;
    }
	
    return 0;
}
