/********************************************************************************
* Copyright (c) 2004-2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*   * Redistributions of source code must retain the above copyright notice,
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its
*     contributors may be used to endorse or promote products derived from
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.

* Author: Hejun Wu (vmnet@ust.hk).
********************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>

#ifndef win32
#ifndef _O_BINARY
#define _O_BINARY 0
#endif
#ifndef O_BINARY
#define O_BINARY _O_BINARY
#endif
#define _O_RDWR O_RDWR
#define _O_CREAT O_CREAT
#define _O_CREAT O_CREAT
#define _O_TRUNC O_TRUNC
#define _S_IWRITE S_IWRITE
#define _S_IREAD S_IREAD 
#define _open open
#define _read read
#define _write write
#define _lseek lseek
#define _tell tell
#else
#include <io.h>
#endif

#define CPU_SLEEP 0x0100
#define CPU_ACTIVE  0x0200

#define ADC_SLEEP  0x0010
#define ADC_CONTROL  0x0020
#define ADC_SAMPLE  0x0040
#define ADC_IDLE    0x0080

#define RADIO_CONTROL 0
#define RADIO_SLEEP 0x0001
#define RADIO_IDLE  0x0002
#define RADIO_RECEIVE  0x0004
#define RADIO_TRANSMIT  0x0008

#define CPU_MASK  0x00ff
#define ADC_MASK  0x0f0f
#define RADIO_MASK 0x0ff0
#ifdef BISQUE_TEST
#define TOTAL_TYPE_RECLOGS 3
#else
#define TOTAL_TYPE_RECLOGS 2
#endif
#define MAXIMUMAL_MODE_NUM 300
unsigned short g_modes[]={141,144,181,182,184,188,202,211,212,221,222,224,241,244,281,282,284,288};


double electricCurrents[300];
char  g_tempMode[10]={0};

//-------------------------------------------------------------------------------
// InitializeCurrents()
// Date: 2004-10-1
// Function: Initialize the electric current in operations.
//-------------------------------------------------------------------------------
void InitializeCurrents()
{
	//Maybe some modes do not occur

	electricCurrents[142]=17.2;
	electricCurrents[144]=25;
	electricCurrents[181]=17.2;
	electricCurrents[182]=17.2;
	electricCurrents[184]=25;
	electricCurrents[188]=28.5;

	electricCurrents[210]=21.2;
	electricCurrents[211]=17.2;
	electricCurrents[220]=21.2;
	electricCurrents[221]=17.2;
	electricCurrents[222]=17.2;
	electricCurrents[224]=25;
	electricCurrents[242]=17.2;
	electricCurrents[244]=25;
	electricCurrents[281]=17.3;
	electricCurrents[280]=21.3;
	electricCurrents[282]=17.3;
	electricCurrents[284]=25;
	electricCurrents[288]=28.5;
}


char * MapToStrMode(int iMode) 
{ 
	char *tempName=g_tempMode;
	if(iMode/100==1)
		sprintf(tempName," %s","I");
	else 
		sprintf(tempName," %s","C");
	tempName+=2;
	switch(iMode%100/10)
	{
	case 1: 
		sprintf(tempName,"%s","P");
		tempName++; 
		break;
	case 2:
	case 8:
		sprintf(tempName,"%s","I");
		tempName++;
		break;
	case 4:
		sprintf(tempName,"%s","A");
		tempName++;
		break;
	default:
		sprintf(tempName,"%s","P");
		tempName++;
		break;
	}
	
	switch(iMode%10)
	{
	case 0:
		sprintf(tempName,"%s","C");
		tempName++;
		break;
	case 1:
		sprintf(tempName,"%s","P");
		tempName++;
		break;
	case 2:
		sprintf(tempName,"%s","I");
		tempName++;
		break;
	case 4:
		sprintf(tempName,"%s","R");
		tempName++;
		break;
	case 8:
		sprintf(tempName,"%s","T");
		tempName++;
		break;
	default:
		sprintf(tempName,"%s","P");
		tempName++;
		break;
	}
	return g_tempMode;
}

typedef struct tagStatus
{
	unsigned short mode;
	struct tagStatus * prevPtr;
	struct tagStatus * nextPtr;
}STAUS_LINK;
unsigned long subtraction_long(unsigned long a, unsigned long b)
{
	unsigned long c;
	if(a>=b)
		c=a-b;
	else
	{
		c=a+((unsigned long)0xffffffffL-b);
		
	}
	return c;
}
void main_task(char* powerLogFileName, int nMoteID,int nRecCounts)
{
    int fh, nFh[TOTAL_TYPE_RECLOGS];
	long i,j=0,k=0,filelen;
	unsigned char *readBuffer;
	char tempName[100]={0}, tempchar[20];
	long durationModes[MAXIMUMAL_MODE_NUM],tempNum, preTime, currentTime=0L, firstTime=0L,totalTime=0L;
	float percentages[MAXIMUMAL_MODE_NUM];
	unsigned short preMode, currentMode=0;
	static short mapMode=0;
	double powerConsumption[MAXIMUMAL_MODE_NUM];
	STAUS_LINK modeLink[MAXIMUMAL_MODE_NUM];
	
	for(i=0;i<MAXIMUMAL_MODE_NUM;i++)
	{
		durationModes[i]=0L;
		modeLink[i].prevPtr=NULL;
		modeLink[i].nextPtr=NULL;
		powerConsumption[i]=0;
	} 
	fh = _open(powerLogFileName,_O_RDWR|_O_BINARY, _S_IREAD 
		| _S_IWRITE );  
	if(fh==-1)
	{
		//printf("No this node ID: %d \n", nMoteID);
		// no this node, try others...
		return;
	}
	filelen=_lseek(fh,0,2);
	readBuffer=malloc(filelen+1);
	_lseek(fh,0,0);
	_read(fh, readBuffer, filelen);
	readBuffer[filelen] =0;
	close(fh);
	
	for (i=1;i<TOTAL_TYPE_RECLOGS;i++)
	{
		memset(tempName,0,80);
		sprintf(tempName,"%d_%d_reclog_%d.txt",nMoteID,nRecCounts,i);
		nFh[i]= _open(tempName,_O_RDWR| _O_CREAT|_O_BINARY|_O_TRUNC, _S_IREAD| _S_IWRITE );  
	
	}
	j=0; 
	while(j<filelen)
	{
		char *tempPtr; 
		while(!((readBuffer[j]>='0'&&readBuffer[j]<='9')||readBuffer[j]=='-')&&j<filelen)
			j++;
		tempPtr=readBuffer +j;
		while(((readBuffer[j]>='0'&&readBuffer[j]<='9')||readBuffer[j]=='-')&&j<filelen)		
			j++;
		if(j<filelen)
			readBuffer[j]=0;
		else 
			break;

		tempNum=atoi(tempPtr);

		if(tempNum>100 && tempNum<300)
		{
			preMode=currentMode;
			currentMode = (unsigned short)tempNum;
			//idle is same as control.Some sensors may be different.
			if((currentMode%100/10)==2)
			{
				currentMode+=60; //idle is same as control.Some sensors may be different.
			}
			if(currentMode/100==2&&currentMode%10==2) currentMode-=2;

			if(currentMode%10==1)currentMode++;
			//if(currentMode==281)currentMode=282;
		}
		else
		{
			preTime = currentTime;
			currentTime = tempNum;
			if(!firstTime)
				firstTime=currentTime;
			if(preMode&& currentTime)
			{
				STAUS_LINK * tempPtr;
				char found=0;
				
				durationModes[preMode]+=subtraction_long(currentTime,preTime);
				totalTime+=subtraction_long(currentTime,preTime);
				
				//for state transition
				#ifdef BISQUE_TEST
				tempPtr=modeLink[preMode].nextPtr;
				while(tempPtr)
				{
					if(tempPtr->mode==currentMode)
					{
						found=1;
						break;
					}
					if(tempPtr->nextPtr)
						tempPtr=tempPtr->nextPtr;
					else
						break;
				}
				if(!tempPtr)
					tempPtr=&modeLink[preMode];
				if(!found)
				{
					tempPtr->nextPtr=malloc(sizeof(STAUS_LINK));
					tempPtr->nextPtr->prevPtr=NULL;
					tempPtr=tempPtr->nextPtr;
					tempPtr->mode=currentMode;
					tempPtr->nextPtr = NULL;
				}
				#endif
				
			}
		}
	}

	for (i=0;i<MAXIMUMAL_MODE_NUM;i++)
	{
		if(durationModes[i])
			percentages[i]=(float)((double)durationModes[i]/(double)totalTime);
		
	}
	for (i=0;i<MAXIMUMAL_MODE_NUM;i++)
	{
		if(durationModes[i])
		{
			memset(tempName,0,100);
			powerConsumption[i]= (double)durationModes[i]*3/(double)7372800*electricCurrents[i]*0.001;
			powerConsumption[0]+=powerConsumption[i];
			
		#ifdef BISQUE_TEST
			
			sprintf(tempName,"%d ",i);	
			_write( nFh[1],tempName, strlen(tempName)); 
			
			sprintf(tempName,"%s",MapToStrMode(i)); 
			_write( nFh[1],tempName, strlen(tempName)); 
			
			sprintf(tempName,"  %.16f ",percentages[i]);
			_write( nFh[1],tempName, strlen(tempName)); 
			
			sprintf(tempName,"  %5.16f ",(double)durationModes[i]/(double)7372.8);
			_write( nFh[1],tempName, strlen(tempName)); 
			
			
			
			sprintf(tempName,"  %5.16f ",electricCurrents[i]);
			_write( nFh[1],tempName, strlen(tempName)); 
			
			sprintf(tempName,"  %5.16f ",powerConsumption[i]);
			_write( nFh[1],tempName, strlen(tempName)); 
			
			mapMode++;

			memset(tempName,0,100);

			sprintf(tempName,"%d\n",durationModes[i]);	

			_write( nFh[1],tempName, strlen(tempName));
			
			if(modeLink[i].prevPtr)
			{
				STAUS_LINK * tempPtr;
				sprintf( tempName,"%s PREV: ", MapToStrMode(i));
				_write( nFh[2],tempName, strlen(tempName));
				tempPtr=modeLink[i].prevPtr;
				while(tempPtr)
				{
					sprintf(tempName,"%s, ",MapToStrMode(tempPtr->mode)); 
					tempPtr = tempPtr->prevPtr;
					_write( nFh[2],tempName, strlen(tempName));
				}
				_write( nFh[2],"\n", 1);
			}
			if((modeLink[i].nextPtr))
			{
				STAUS_LINK * tempPtr;
				
				sprintf( tempName," %s NEXT: ", MapToStrMode(i));
				_write( nFh[2],tempName, strlen(tempName));
				tempPtr=modeLink[i].nextPtr;
				while(tempPtr)
				{
					sprintf(tempName,"%s, ",MapToStrMode(tempPtr->mode)); 
					tempPtr = tempPtr->nextPtr;
					_write( nFh[2],tempName, strlen(tempName));
				}
				_write( nFh[2],"\n", 1);
			}
			#endif
			
		}
		
	}
	sprintf(tempName,"\t\t\t%5.16f ",(double)totalTime/(double)7372.8);
	_write( nFh[1],tempName, strlen(tempName)); 
	sprintf(tempName,"\t\t%5.16f ",powerConsumption[0]);
	_write( nFh[1],tempName, strlen(tempName));
	
	
	for (i=0;i<TOTAL_TYPE_RECLOGS;i++)
		close(nFh[i]);
	free(readBuffer);
	
}

#define EXP_TIMES 200
#define NODE_NUM  100
int main()
{
	int i,j;
	char tempName[100]={0};
	InitializeCurrents();
	// analyze the logs node by node
	for (i=0;i<NODE_NUM;i++)
	{
		for (j=0;j<EXP_TIMES;j++)		
		{
			sprintf(tempName,"power_log_%d_%d.rec",i,j);
			main_task(tempName,i,j);
		}
	}
	return 0;
}   


